# -*- coding: utf-8 -*-
'''
SSH Proxy
=========

Manage a remote host via SSH, using a Proxy Minion. This module doesn't have any
external dependencies, as it makes use of the native Salt internals used for
salt-ssh, therefore managing the remote machine by uploading a lightweight Salt
version on the target host, then invokes Salt functions over SSH (using the
``ssh`` binary installed on your computer or wherever this Proxy Minion runs).

.. note::

    To manage machines running Windows, you will need to install the
    ``saltwinshell`` library.

Pillar
------

The configuration is aligned to the general Proxy Minion standards: put the
connection details and credentials under the ``proxy`` key in the Proxy config
or Pillar.

.. important:

    Local (i.e., per Proxy) option override the global configuration or CLI
    options.

``host``
    The IP address or the hostname of the remove machine to manage.

``port``
    Integer, the port number to use when establishing he connection
    (defaults to 22).

``user``
    The username required for authentication.

``passwd``
    The password used for authentication.

``priv``
    Absolute path to the private SSH key used for authentication.

``priv_passwd``
    The SSH private key password.

``timeout``: 30
    The SSH timeout. Defaults to 30 seconds.

``sudo``: ``False``
    Execute commands as sudo.

``tty``: ``False``
    Connect over tty.

``sudo_user``
    The username that should execute the commands as sudo.

``remote_port_forwards``
    Enable remote port forwarding. Example: ``8888:my.company.server:443``.
    Multiple remote port forwardings are supported, using comma-separated
    values, e.g., ``8888:my.company.server:443,9999:my.company.server:80``.

``identities_only``: ``False``
    Execute SSH with ``-o IdentitiesOnly=yes``. This option is intended for
    situations where ssh-agent offers many different identities and allow ssh
    to ignore those identities and use the only one specified in options.

``ignore_host_keys``: ``False``
    By default ssh host keys are honored and connections will ask for approval.
    Use this option to disable ``StrictHostKeyChecking``.

``no_host_keys``: ``False``
    Fully ignores ssh host keys which by default are honored and connections
    would ask for approval. Useful if the host key of a remote server has
    changed and would still error with ``ignore_host_keys``.

``winrm``: ``False``
    Flag that tells Salt to connect to a Windows machine. This option requires
    the ``saltwinshell`` to be installed.

Example Pillar:

.. code-block:: yaml

  proxy:
    proxytype: ssh
    host: srv.example.com
    user: test
    passwd: test
    port: 2022
'''
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging

import salt.client.ssh
import salt.fileclient
import salt.exceptions
import salt.utils.path
from salt.ext import six

__proxyenabled__ = ['ssh']

log = logging.getLogger(__name__)

CONN = None
INITIALIZED = False
GRAINS_CACHE = {}


def _prep_conn(opts, fun, *args, **kwargs):
    '''
    Prepare the connection.
    '''
    opts['_ssh_version'] = salt.client.ssh.ssh_version()
    fsclient = salt.fileclient.FSClient(opts)
    # TODO: Have here more options to simplify the usage, through features like
    # auto-expand the path to the priv key, auto-discovery, etc.
    argv = [fun]
    argv.extend([salt.utils.json.dumps(arg) for arg in args])
    argv.extend(
        [
            "{0}={1}".format(
                salt.utils.stringutils.to_str(key), salt.utils.json.dumps(val)
            )
            for key, val in six.iteritems(kwargs)
        ]
    )
    if not opts['proxy'].get('ssh_options'):
        opts['proxy']['ssh_options'] = []
    if opts['proxy'].get('ignore_host_keys', False):
        opts['proxy']['ssh_options'].append('StrictHostKeyChecking=no')
    if opts['proxy'].get('no_host_keys', False):
        opts['proxy']['ssh_options'].extend(
            ["StrictHostKeyChecking=no", "UserKnownHostsFile=/dev/null"]
        )
    for cli_opt in ('identities_only', 'priv', 'priv_passwd'):
        if opts.get(cli_opt) and not opts['proxy'].get(cli_opt):
            opts['proxy'][cli_opt] = opts[cli_opt]
    ext_mods = salt.client.ssh.mod_data(fsclient)
    conn = salt.client.ssh.Single(
        opts, argv, opts['id'], fsclient=fsclient, mods=ext_mods, **opts['proxy']
    )
    conn.args = args
    conn.kwargs = kwargs
    thin_dir = conn.opts['thin_dir']
    thin_dir = thin_dir.replace('proxy', '')
    conn.opts['thin_dir'] = thin_dir
    conn.thin_dir = thin_dir
    return conn


def init(opts):
    '''
    Init the SSH connection, and execute a simple call to ensure that the remote
    device is reachable, otherwise throw an error.
    '''
    global CONN, INITIALIZED
    if not salt.utils.path.which('ssh'):
        raise salt.exceptions.SaltSystemExit(
            code=-1,
            msg='No ssh binary found in path -- ssh must be installed for this Proxy module. Exiting.',
        )
    CONN = _prep_conn(opts, 'cmd.run', 'echo')
    INITIALIZED = True


def initialized():
    '''
    Proxy initialized properly?
    '''
    return INITIALIZED


def module_executors():
    '''
    Return the list of executors that should invoke the Salt functions.
    '''
    return ['ssh']


def call(fun, *args, **kwargs):
    '''
    Call an arbitrary Salt function and return the output.
    '''
    global CONN, INITIALIZED
    if not CONN or not INITIALIZED:
        return
    opts = CONN.opts
    opts['output'] = 'json'
    ssh_conn = _prep_conn(opts, fun, *args, **kwargs)
    ret = ssh_conn.run()
    if ret[2] != 0:
        log.error('[%s] %s', opts['id'], ret[1])
        return ret[0]
    thin_ret = json.loads(ret[0])
    if '_error' in thin_ret['local']:
        log.error(thin_ret['local']['_error'])
        if 'stdout' in thin_ret['local']:
            log.error(thin_ret['local']['stdout'])
    return thin_ret['local']['return']


def ping():
    '''
    Execute "echo" on the remote host to ensure it's still accessible.
    '''
    global CONN, INITIALIZED
    if not CONN or not INITIALIZED:
        log.debug('Not connected, or not initialized')
        return False
    ret = CONN.run()
    log.debug(ret)
    return ret[2] == 0


def grains():
    '''
    Invoke grains.items from the thin Salt on the remote machine, in order to
    return here the Grains.
    '''
    global GRAINS_CACHE
    if not GRAINS_CACHE:
        GRAINS_CACHE = call('grains.items')
    return GRAINS_CACHE


def shutdown(opts):
    '''
    Buh-bye...
    '''
    global CONN, INITIALIZED
    if CONN and INITIALIZED:
        del CONN
        INITIALIZED = False



# -*- coding: utf-8 -*-
'''
Read in an Ansible inventory file or script

Flat inventory files should be in the regular ansible inventory format.

.. code-block:: ini

    [servers]
    salt.gtmanfred.com ansible_ssh_user=gtmanfred ansible_ssh_host=127.0.0.1 ansible_ssh_port=22 ansible_ssh_pass='password'

    [desktop]
    home ansible_ssh_user=gtmanfred ansible_ssh_host=12.34.56.78 ansible_ssh_port=23 ansible_ssh_pass='password'

    [computers:children]
    desktop
    servers

    [names:vars]
    http_port=80

then salt-ssh can be used to hit any of them

.. code-block:: bash

    [~]# salt-ssh -N all test.ping
    salt.gtmanfred.com:
        True
    home:
        True
    [~]# salt-ssh -N desktop test.ping
    home:
        True
    [~]# salt-ssh -N computers test.ping
    salt.gtmanfred.com:
        True
    home:
        True
    [~]# salt-ssh salt.gtmanfred.com test.ping
    salt.gtmanfred.com:
        True

There is also the option of specifying a dynamic inventory, and generating it on the fly

.. code-block:: bash

    #!/bin/bash
    echo '{
      "servers": [
        "salt.gtmanfred.com"
      ],
      "desktop": [
        "home"
      ],
      "computers": {
        "hosts": [],
        "children": [
          "desktop",
          "servers"
        ]
      },
      "_meta": {
        "hostvars": {
          "salt.gtmanfred.com": {
            "ansible_ssh_user": "gtmanfred",
            "ansible_ssh_host": "127.0.0.1",
            "ansible_sudo_pass": "password",
            "ansible_ssh_port": 22
          },
          "home": {
            "ansible_ssh_user": "gtmanfred",
            "ansible_ssh_host": "12.34.56.78",
            "ansible_sudo_pass": "password",
            "ansible_ssh_port": 23
          }
        }
      }
    }'

This is the format that an inventory script needs to output to work with ansible, and thus here.

.. code-block:: bash

    [~]# salt-ssh --roster-file /etc/salt/hosts salt.gtmanfred.com test.ping
    salt.gtmanfred.com:
            True

Any of the [groups] or direct hostnames will return.  The 'all' is special, and returns everything.
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import copy
import json
import logging

# Import Salt libs
from salt.roster import get_roster_file

try:
    from salt.utils.path import which as utils_which
    from salt.utils.stringutils import to_str as utils_to_str
except ImportError:
    from salt.utils import which as utils_which
    from salt.utils import to_str as utils_to_str

# Import salt-sproxy modules
import salt_sproxy._roster

log = logging.getLogger(__name__)

CONVERSION = {
    'ansible_ssh_host': 'host',
    'ansible_ssh_port': 'port',
    'ansible_ssh_user': 'user',
    'ansible_ssh_pass': 'passwd',
    'ansible_sudo_pass': 'sudo',
    'ansible_ssh_private_key_file': 'priv',
}

__virtualname__ = 'ansible'


def __virtual__():
    return (
        utils_which('ansible-inventory') and __virtualname__,
        'Install `ansible` to use inventory',
    )


def targets(tgt, tgt_type='glob', **kwargs):
    '''
    Return the targets from the ansible inventory_file
    Default: /etc/salt/roster
    '''
    inventory = __runner__['salt.cmd'](
        'cmd.run', 'ansible-inventory -i {0} --list'.format(get_roster_file(__opts__))
    )
    __context__['inventory'] = json.loads(utils_to_str(inventory))

    if tgt_type == 'nodegroup':
        hosts = _get_hosts_from_group(tgt)
        return {host: _get_hostvars(host) for host in hosts}
    pool = {host: _get_hostvars(host) for host in _get_hosts_from_group('all')}
    pool = salt_sproxy._roster.load_cache(
        pool, __runner__, __opts__, tgt, tgt_type=tgt_type
    )
    log.debug('Ansible devices pool')
    log.debug(pool)
    engine = salt_sproxy._roster.TGT_FUN[tgt_type]
    return engine(pool, tgt, opts=__opts__)


def _get_hosts_from_group(group):
    inventory = __context__['inventory']
    hosts = [host for host in inventory.get(group, {}).get('hosts', [])]
    for child in inventory.get(group, {}).get('children', []):
        hosts.extend(_get_hosts_from_group(child))
    return hosts


def _get_hostvars(host):
    hostvars = __context__['inventory']['_meta'].get('hostvars', {}).get(host, {})
    ret = copy.deepcopy(__opts__.get('roster_defaults', {}))
    for value in CONVERSION:
        if value in hostvars:
            ret[CONVERSION[value]] = hostvars.pop(value)
    ret['minion_opts'] = hostvars
    if 'host' not in ret:
        ret['host'] = host
    return ret



# -*- coding: utf-8 -*-
# Copyright 2019-2020 Mircea Ulinic. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
'''
Salt Runner to invoke arbitrary commands on network devices that are not
managed via a Proxy or regular Minion. Therefore, this Runner doesn't
necessarily require the targets to be up and running, as it will connect to
collect the Grains, compile the Pillar, then execute the commands.
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python std lib
import sys
import copy
import json
import math
import time
import hashlib
import logging
import threading
import traceback
import multiprocessing

# Import Salt modules
import salt.cache
import salt.loader
import salt.client
import salt.output
import salt.version
import salt.utils.jid
import salt.utils.master
from salt.ext import six
from salt.minion import SMinion
from salt.cli.batch import Batch
import salt.utils.stringutils
import salt.defaults.exitcodes
from salt.exceptions import SaltSystemExit, SaltInvocationError
from salt.defaults import DEFAULT_TARGET_DELIM

import salt.utils.napalm
import salt.utils.dictupdate

try:
    import salt.utils.platform
    from salt.utils.args import clean_kwargs

    OLD_SALT = False
except ImportError:
    OLD_SALT = True
    import salt.utils
    from salt.utils import clean_kwargs

try:
    import progressbar

    HAS_PROGRESSBAR = True
except ImportError:
    HAS_PROGRESSBAR = False

# ------------------------------------------------------------------------------
# module properties
# ------------------------------------------------------------------------------

_SENTINEL = 'FIN.'

log = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
# property functions
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------


def _napalm_is_proxy(opts):
    return opts.get('proxy', {}).get('proxytype') == 'napalm'


# Point the native is_proxy function to the above, so it doesn't check whether
# we're actually running under a Proxy Minion
salt.utils.napalm.is_proxy = _napalm_is_proxy


def _is_proxy():
    return True


# Same rationale as above, for any other Proxy type.
if not OLD_SALT:
    salt.utils.platform.is_proxy = _is_proxy
else:
    salt.utils.is_proxy = _is_proxy


def _salt_call_and_return(
    minion_id,
    salt_function,
    ret_queue,
    unreachable_devices,
    failed_devices,
    arg=None,
    jid=None,
    events=True,
    **opts
):
    '''
    '''
    opts['jid'] = jid
    ret, retcode = salt_call(
        minion_id,
        salt_function,
        unreachable_devices=unreachable_devices,
        failed_devices=failed_devices,
        **opts
    )
    if events:
        __salt__['event.send'](
            'proxy/runner/{jid}/ret/{minion_id}'.format(minion_id=minion_id, jid=jid),
            {
                'fun': salt_function,
                'fun_args': arg,
                'id': minion_id,
                'jid': jid,
                'return': ret,
                'retcode': retcode,
                'success': retcode == 0,
            },
        )
    try:
        ret = json.loads(json.dumps(ret))
    except (ValueError, TypeError):
        log.error('Function return is not JSON-serializable data', exc_info=True)
        log.error(ret)
    ret_queue.put(({minion_id: ret}, retcode))
    sys.exit(retcode)


def _existing_proxy_cli_batch(
    cli_batch, ret_queue, batch_stop_queue, sproxy_stop_queue
):
    '''
    '''
    run = cli_batch.run()
    cumulative_retcode = 0
    for ret in run:
        if not sproxy_stop_queue.empty():
            break
        retcode = 0
        if ret and isinstance(ret, dict):
            minion_id = list(ret.keys())[0]
            if isinstance(ret[minion_id], dict) and 'retcode' in ret[minion_id]:
                retcode = ret[minion_id].pop('retcode')
        ret_queue.put((ret, retcode))
        cumulative_retcode = max(cumulative_retcode, retcode)
    batch_stop_queue.put(cumulative_retcode)


def _receive_replies_async(ret_queue, done_queue, progress_bar):
    '''
    '''
    count = 0
    while True:
        ret, retcode = ret_queue.get()
        count += 1
        if ret == _SENTINEL:
            break
        # When async, print out the replies as soon as they arrive
        # after passing them through the outputter of choice
        out_fmt = salt.output.out_format(
            ret, __opts__.get('output', 'nested'), opts=__opts__, _retcode=retcode,
        )
        if out_fmt:
            # out_fmt can be empty string, for example, when using the ``quiet``
            # outputter, or potentially other use cases.
            salt.utils.stringutils.print_cli(out_fmt)
        if progress_bar:
            progress_bar.update(count)
    done_queue.put(_SENTINEL)


def _receive_replies_sync(ret_queue, static_queue, done_queue, progress_bar):
    '''
    '''
    count = 0
    cumulative_retcode = 0
    while True:
        ret, retcode = ret_queue.get()
        static_queue.put((ret, retcode))
        count += 1
        if ret == _SENTINEL:
            break
        if progress_bar:
            progress_bar.update(count)
    done_queue.put(_SENTINEL)


class NoPingBatch(Batch):
    '''
    Similar to the native Salt Batch. but without issuing test.ping to ensure
    that the Minions are up and running.
    '''

    def __init__(
        self, opts, eauth=None, quiet=False, parser=None
    ):  # pylint: disable=super-init-not-called
        self.opts = opts
        self.eauth = eauth if eauth else {}
        self.pub_kwargs = eauth if eauth else {}
        self.quiet = quiet
        self.local = salt.client.get_local_client(opts['conf_file'])
        self.minions, self.ping_gen, self.down_minions = self.__gather_minions()
        self.options = parser

    def __gather_minions(self):
        return self.opts['tgt'], [], []


# The SProxyMinion class is back-ported from Salt 2019.2.0 (to be released soon)
# and extended to allow more flexible options for the (pre-)loading of the
# Pillars and the Grains.
class SProxyMinion(SMinion):
    '''
    Create an object that has loaded all of the minion module functions,
    grains, modules, returners etc.  The SProxyMinion allows developers to
    generate all of the salt minion functions and present them with these
    functions for general use.
    '''

    def _matches_target(self):
        match_func = self.matchers.get(
            '{0}_match.match'.format(self.opts['__tgt_type']), None
        )
        if match_func is None:
            return False
        if self.opts['__tgt_type'] in ('grain', 'grain_pcre', 'pillar'):
            delimiter = self.opts.get('delimiter', DEFAULT_TARGET_DELIM)
            if not match_func(self.opts['__tgt'], delimiter=delimiter):
                return False
        elif not match_func(self.opts['__tgt']):
            return False
        else:
            if not self.matchers['glob_match.match'](self.opts['__tgt']):
                return False
        return True

    def gen_modules(self, initial_load=False):  # pylint: disable=arguments-differ
        '''
        Tell the minion to reload the execution modules.

        CLI Example:

        .. code-block:: bash

            salt '*' sys.reload_modules
        '''
        if self.opts.get('proxy_preload_grains', True):
            loaded_grains = salt.loader.grains(self.opts)
            self.opts['grains'].update(loaded_grains)

        if (
            self.opts['roster_opts']
            and self.opts.get('proxy_merge_roster_grains', True)
            and 'grains' in self.opts['roster_opts']
            and isinstance(self.opts['roster_opts']['grains'], dict)
        ):
            # Merge the Grains from the Roster opts
            log.debug('Merging Grains with the Roster provided ones')
            self.opts['grains'] = salt.utils.dictupdate.merge(
                self.opts['roster_opts']['grains'], self.opts['grains']
            )

        cached_grains = None
        if self.opts.get('proxy_use_cached_grains', True):
            cached_grains = self.opts.pop('proxy_cached_grains', None)

        initial_grains = copy.deepcopy(self.opts['grains'])
        if cached_grains:
            # Merging the collected Grains into the cached Grains, but only for
            # the initial Pillar compilation, to ensure we only do so to avoid
            # any processing errors.
            initial_grains = salt.utils.dictupdate.merge(cached_grains, initial_grains)

        if self.opts.get('proxy_load_pillar', True):
            self.opts['pillar'] = salt.pillar.get_pillar(
                self.opts,
                initial_grains,
                self.opts['id'],
                saltenv=self.opts['saltenv'],
                pillarenv=self.opts.get('pillarenv'),
            ).compile_pillar()

        if self.opts['roster_opts'] and self.opts.get('proxy_merge_roster_opts', True):
            if 'proxy' not in self.opts['pillar']:
                self.opts['pillar']['proxy'] = {}
            self.opts['pillar']['proxy'] = salt.utils.dictupdate.merge(
                self.opts['pillar']['proxy'], self.opts['roster_opts']
            )
            self.opts['pillar']['proxy'].pop('grains', None)
            self.opts['pillar']['proxy'].pop('pillar', None)

        if self.opts.get('preload_targeting', False) or self.opts.get(
            'invasive_targeting', False
        ):
            log.debug('Loading the Matchers modules')
            self.matchers = salt.loader.matchers(self.opts)

        if self.opts.get('preload_targeting', False):
            log.debug(
                'Preload targeting requested, trying to see if %s matches the target %s (%s)',
                self.opts['id'],
                str(self.opts['__tgt']),
                self.opts['__tgt_type'],
            )
            matched = self._matches_target()
            if not matched:
                return

        if 'proxy' not in self.opts['pillar'] and 'proxy' not in self.opts:
            errmsg = (
                'No "proxy" configuration key found in pillar or opts '
                'dictionaries for id {id}. Check your pillar/options '
                'configuration and contents. Salt-proxy aborted.'
            ).format(id=self.opts['id'])
            log.error(errmsg)
            self._running = False
            raise SaltSystemExit(code=salt.defaults.exitcodes.EX_GENERIC, msg=errmsg)

        if 'proxy' not in self.opts:
            self.opts['proxy'] = {}
        if 'proxy' not in self.opts['pillar']:
            self.opts['pillar']['proxy'] = {}
        self.opts['proxy'] = salt.utils.dictupdate.merge(
            self.opts['proxy'], self.opts['pillar']['proxy']
        )

        # Then load the proxy module
        fq_proxyname = self.opts['proxy']['proxytype']
        self.utils = salt.loader.utils(self.opts)
        self.proxy = salt.loader.proxy(
            self.opts, utils=self.utils, whitelist=[fq_proxyname]
        )
        self.functions = salt.loader.minion_mods(
            self.opts, utils=self.utils, notify=False, proxy=self.proxy
        )
        self.functions.pack['__grains__'] = copy.deepcopy(self.opts['grains'])

        self.functions.pack['__proxy__'] = self.proxy
        self.proxy.pack['__salt__'] = self.functions
        self.proxy.pack['__pillar__'] = self.opts['pillar']

        # No need to inject the proxy into utils, as we don't need scheduler for
        # this sort of short living Minion.
        # self.utils = salt.loader.utils(self.opts, proxy=self.proxy)
        self.proxy.pack['__utils__'] = self.utils

        # Reload all modules so all dunder variables are injected
        self.proxy.reload_modules()

        if self.opts.get('proxy_no_connect', False):
            log.info('Requested not to initialize the connection with the device')
        else:
            log.debug('Trying to initialize the connection with the device')
            # When requested --no-connect, don't init the connection, but simply
            # go ahead and execute the function requested.
            if (
                '{0}.init'.format(fq_proxyname) not in self.proxy
                or '{0}.shutdown'.format(fq_proxyname) not in self.proxy
            ):
                errmsg = (
                    '[{0}] Proxymodule {1} is missing an init() or a shutdown() or both. '.format(
                        self.opts['id'], fq_proxyname
                    )
                    + 'Check your proxymodule.  Salt-proxy aborted.'
                )
                log.error(errmsg)
                self._running = False
                if self.unreachable_devices is not None:
                    self.unreachable_devices.append(self.opts['id'])
                raise SaltSystemExit(
                    code=salt.defaults.exitcodes.EX_GENERIC, msg=errmsg
                )

            proxy_init_fn = self.proxy[fq_proxyname + '.init']
            try:
                proxy_init_fn(self.opts)
                self.connected = True
            except Exception as exc:
                log.error(
                    'Encountered error when starting up the connection with %s:',
                    self.opts['id'],
                    exc_info=True,
                )
                if self.unreachable_devices is not None:
                    self.unreachable_devices.append(self.opts['id'])
                raise
            if self.opts.get('proxy_load_grains', True):
                # When the Grains are loaded from the cache, no need to re-load them
                # again.

                grains = copy.deepcopy(self.opts['grains'])
                # Copy the existing Grains loaded so far, otherwise
                # salt.loader.grains is going to wipe what's under the grains
                # key in the opts.
                # After loading, merge with the previous loaded grains, which
                # may contain other grains from different sources, e.g., roster.
                loaded_grains = salt.loader.grains(self.opts, proxy=self.proxy)
                self.opts['grains'] = salt.utils.dictupdate.merge(grains, loaded_grains)
            if self.opts.get('proxy_load_pillar', True):
                self.opts['pillar'] = salt.pillar.get_pillar(
                    self.opts,
                    self.opts['grains'],
                    self.opts['id'],
                    saltenv=self.opts['saltenv'],
                    pillarenv=self.opts.get('pillarenv'),
                ).compile_pillar()
            self.functions.pack['__opts__'] = self.opts
            self.functions.pack['__grains__'] = copy.deepcopy(self.opts['grains'])
            self.functions.pack['__pillar__'] = copy.deepcopy(self.opts['pillar'])
        self.grains_cache = copy.deepcopy(self.opts['grains'])

        if self.opts.get('invasive_targeting', False):
            log.info(
                'Invasive targeting requested, trying to see if %s matches the target %s (%s)',
                self.opts['id'],
                str(self.opts['__tgt']),
                self.opts['__tgt_type'],
            )
            matched = self._matches_target()
            if not matched:
                # Didn't match, shutting down this Proxy Minion, and exiting.
                log.debug(
                    '%s does not match the target expression, aborting', self.opts['id']
                )
                proxy_shut_fn = self.proxy[fq_proxyname + '.shutdown']
                proxy_shut_fn(self.opts)
                return

        self.module_executors = self.proxy.get(
            '{0}.module_executors'.format(fq_proxyname), lambda: []
        )() or self.opts.get('module_executors', [])
        if self.module_executors:
            self.executors = salt.loader.executors(
                self.opts, self.functions, proxy=self.proxy
            )

        # Late load the Returners, as they might need Grains, which may not be
        # properly or completely loaded before this.
        self.returners = None
        if self.opts['returner']:
            self.returners = salt.loader.returners(
                self.opts, self.functions, proxy=self.proxy
            )
        self.proxy.pack['__ret__'] = self.returners

        self.ready = True


class StandaloneProxy(SProxyMinion):
    def __init__(
        self, opts, unreachable_devices=None
    ):  # pylint: disable=super-init-not-called
        self.opts = opts
        self.connected = False
        self.ready = False
        self.unreachable_devices = unreachable_devices
        self.gen_modules()


# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def salt_call(
    minion_id,
    salt_function=None,
    unreachable_devices=None,
    failed_devices=None,
    with_grains=True,
    with_pillar=True,
    preload_grains=True,
    preload_pillar=True,
    default_grains=None,
    default_pillar=None,
    cache_grains=True,
    cache_pillar=True,
    use_cached_grains=True,
    use_cached_pillar=True,
    use_existing_proxy=False,
    no_connect=False,
    jid=None,
    roster_opts=None,
    test_ping=False,
    tgt=None,
    tgt_type=None,
    preload_targeting=False,
    invasive_targeting=False,
    failhard=False,
    timeout=60,
    returner='',
    returner_config='',
    returner_kwargs=None,
    args=(),
    **kwargs
):
    '''
    Invoke a Salt Execution Function that requires or invokes an NAPALM
    functionality (directly or indirectly).

    minion_id:
        The ID of the Minion to compile Pillar data for.

    salt_function
        The name of the Salt function to invoke.

    preload_grains: ``True``
        Whether to preload the Grains before establishing the connection with
        the remote network device.

    default_grains:
        Dictionary of the default Grains to make available within the functions
        loaded.

    with_grains: ``True``
        Whether to load the Grains modules and collect Grains data and make it
        available inside the Execution Functions.
        The Grains will be loaded after opening the connection with the remote
        network device.

    preload_pillar: ``True``
        Whether to preload Pillar data before opening the connection with the
        remote network device.

    default_pillar:
        Dictionary of the default Pillar data to make it available within the
        functions loaded.

    with_pillar: ``True``
        Whether to load the Pillar modules and compile Pillar data and make it
        available inside the Execution Functions.

    use_cached_pillar: ``True``
        Use cached Pillars whenever possible. If unable to gather cached data,
        it falls back to compiling the Pillar.

    use_cached_grains: ``True``
        Use cached Grains whenever possible. If unable to gather cached data,
        it falls back to collecting Grains.

    cache_pillar: ``True``
        Cache the compiled Pillar data before returning.

    cache_grains: ``True``
        Cache the collected Grains before returning.

    use_existing_proxy: ``False``
        Use the existing Proxy Minions when they are available (say on an
        already running Master).

    no_connect: ``False``
        Don't attempt to initiate the connection with the remote device.
        Default: ``False`` (it will initiate the connection).

    jid: ``None``
        The JID to pass on, when executing.

    test_ping: ``False``
        When using the existing Proxy Minion with the ``use_existing_proxy``
        option, can use this argument to verify also if the Minion is
        responsive.

    arg
        The list of arguments to send to the Salt function.

    kwargs
        Key-value arguments to send to the Salt function.

    CLI Example:

    .. code-block:: bash

        salt-run proxy.salt_call bgp.neighbors junos 1.2.3.4 test test123
        salt-run proxy.salt_call net.load_config junos 1.2.3.4 test test123 text='set system ntp peer 1.2.3.4'
    '''
    opts = copy.deepcopy(__opts__)
    opts['id'] = minion_id
    opts['pillarenv'] = __opts__.get('pillarenv', 'base')
    opts['__cli'] = __opts__.get('__cli', 'salt-call')
    opts['__tgt'] = tgt
    opts['__tgt_type'] = tgt_type
    if 'saltenv' not in opts:
        opts['saltenv'] = 'base'
    if not default_grains:
        default_grains = {}
    opts['grains'] = default_grains
    if not default_pillar:
        default_pillar = {}
    opts['pillar'] = default_pillar
    opts['proxy_load_pillar'] = with_pillar
    opts['proxy_load_grains'] = with_grains
    opts['proxy_preload_pillar'] = preload_pillar
    opts['proxy_preload_grains'] = preload_grains
    opts['proxy_cache_grains'] = cache_grains
    opts['proxy_cache_pillar'] = cache_pillar
    opts['preload_targeting'] = preload_targeting
    opts['invasive_targeting'] = invasive_targeting
    opts['proxy_no_connect'] = no_connect
    opts['proxy_test_ping'] = test_ping
    opts['proxy_use_cached_grains'] = use_cached_grains
    if use_cached_grains:
        opts['proxy_cached_grains'] = __salt__['cache.fetch'](
            'minions/{}/data'.format(minion_id), 'grains'
        )
    opts['roster_opts'] = roster_opts
    opts['returner'] = returner
    if not returner_kwargs:
        returner_kwargs = {}
    minion_defaults = salt.config.DEFAULT_MINION_OPTS.copy()
    minion_defaults.update(salt.config.DEFAULT_PROXY_MINION_OPTS)
    for opt, val in six.iteritems(minion_defaults):
        if opt not in opts:
            opts[opt] = val
    sa_proxy = StandaloneProxy(opts, unreachable_devices)
    if not sa_proxy.ready:
        log.debug(
            'The SProxy Minion for %s is not able to start up, aborting', opts['id']
        )
        return
    kwargs = clean_kwargs(**kwargs)
    ret = None
    retcode = 0
    executors = getattr(sa_proxy, 'module_executors')
    try:
        if executors:
            for name in executors:
                ex_name = '{}.execute'.format(name)
                if ex_name not in sa_proxy.executors:
                    raise SaltInvocationError(
                        "Executor '{0}' is not available".format(name)
                    )
                ret = sa_proxy.executors[ex_name](
                    opts, {'fun': salt_function}, salt_function, args, kwargs
                )
                if ret is not None:
                    break
        else:
            ret = sa_proxy.functions[salt_function](*args, **kwargs)
        retcode = sa_proxy.functions.pack['__context__'].get('retcode', 0)
    except Exception as err:
        log.info('Exception while running %s on %s', salt_function, opts['id'])
        if failed_devices is not None:
            failed_devices.append(opts['id'])
        ret = 'The minion function caused an exception: {err}'.format(
            err=traceback.format_exc()
        )
        if not retcode:
            retcode = 11
        if failhard:
            raise
    finally:
        if sa_proxy.connected:
            shut_fun = '{}.shutdown'.format(sa_proxy.opts['proxy']['proxytype'])
            sa_proxy.proxy[shut_fun](opts)
    if returner:
        returner_fun = '{}.returner'.format(returner)
        if returner_fun in sa_proxy.returners:
            log.debug(
                'Sending the response from %s to the %s Returner', opts['id'], returner,
            )
            ret_data = {
                'id': opts['id'],
                'jid': jid,
                'fun': salt_function,
                'fun_args': args,
                'return': ret,
                'ret_config': returner_config,
                'ret_kwargs': returner_kwargs,
            }
            try:
                sa_proxy.returners[returner_fun](ret_data)
            except Exception as err:
                log.error(
                    'Exception while sending the response from %s to the %s returner',
                    opts['id'],
                    returner,
                )
                log.error(err, exc_info=True)
        else:
            log.warning(
                'Returner %s is not available. Check that the dependencies are properly installed'
            )
    if cache_grains:
        log.debug('Caching Grains for %s', minion_id)
        log.debug(sa_proxy.opts['grains'])
        cache_store = __salt__['cache.store'](
            'minions/{}/data'.format(minion_id), 'grains', sa_proxy.opts['grains']
        )
    if cache_pillar:
        log.debug('Caching Pillar for %s', minion_id)
        cached_store = __salt__['cache.store'](
            'minions/{}/data'.format(minion_id), 'pillar', sa_proxy.opts['pillar']
        )
    return ret, retcode


def execute_devices(
    minions,
    salt_function,
    with_grains=True,
    with_pillar=True,
    preload_grains=True,
    preload_pillar=True,
    default_grains=None,
    default_pillar=None,
    args=(),
    batch_size=None,
    batch_wait=0,
    static=False,
    tgt=None,
    tgt_type=None,
    jid=None,
    events=True,
    cache_grains=True,
    cache_pillar=True,
    use_cached_grains=True,
    use_cached_pillar=True,
    use_existing_proxy=False,
    existing_minions=None,
    no_connect=False,
    roster_targets=None,
    test_ping=False,
    preload_targeting=False,
    invasive_targeting=False,
    failhard=False,
    timeout=60,
    summary=False,
    verbose=False,
    progress=False,
    hide_timeout=False,
    returner='',
    returner_config='',
    returner_kwargs=None,
    **kwargs
):
    '''
    Execute a Salt function on a group of network devices identified by their
    Minion ID, as listed under the ``minions`` argument.

    minions
        A list of Minion IDs to invoke ``function`` on.

    salt_function
        The name of the Salt function to invoke.

    preload_grains: ``True``
        Whether to preload the Grains before establishing the connection with
        the remote network device.

    default_grains:
        Dictionary of the default Grains to make available within the functions
        loaded.

    with_grains: ``False``
        Whether to load the Grains modules and collect Grains data and make it
        available inside the Execution Functions.
        The Grains will be loaded after opening the connection with the remote
        network device.

    preload_pillar: ``True``
        Whether to preload Pillar data before opening the connection with the
        remote network device.

    default_pillar:
        Dictionary of the default Pillar data to make it available within the
        functions loaded.

    with_pillar: ``True``
        Whether to load the Pillar modules and compile Pillar data and make it
        available inside the Execution Functions.

    args
        The list of arguments to send to the Salt function.

    kwargs
        Key-value arguments to send to the Salt function.

    batch_size: None
        The size of each batch to execute.

    static: ``False``
        Whether to return the results synchronously (or return them as soon
        as the device replies).

    events: ``True``
        Whether should push events on the Salt bus, similar to when executing
        equivalent through the ``salt`` command.

    use_cached_pillar: ``True``
        Use cached Pillars whenever possible. If unable to gather cached data,
        it falls back to compiling the Pillar.

    use_cached_grains: ``True``
        Use cached Grains whenever possible. If unable to gather cached data,
        it falls back to collecting Grains.

    cache_pillar: ``True``
        Cache the compiled Pillar data before returning.

    cache_grains: ``True``
        Cache the collected Grains before returning.

    use_existing_proxy: ``False``
        Use the existing Proxy Minions when they are available (say on an
        already running Master).

    no_connect: ``False``
        Don't attempt to initiate the connection with the remote device.
        Default: ``False`` (it will initiate the connection).

    test_ping: ``False``
        When using the existing Proxy Minion with the ``use_existing_proxy``
        option, can use this argument to verify also if the Minion is
        responsive.

    CLI Example:

    .. code-block:: bash

        salt-run proxy.execute "['172.17.17.1', '172.17.17.2']" test.ping driver=eos username=test password=test123
    '''
    resp = ''
    retcode = 0
    __pub_user = kwargs.get('__pub_user')
    if not __pub_user:
        __pub_user = __utils__['user.get_specific_user']()
    kwargs = clean_kwargs(**kwargs)
    if not jid:
        if salt.version.__version_info__ >= (2018, 3, 0):
            jid = salt.utils.jid.gen_jid(__opts__)
        else:
            jid = salt.utils.jid.gen_jid()  # pylint: disable=no-value-for-parameter
    event_args = list(args[:])
    if kwargs:
        event_kwargs = {'__kwarg__': True}
        event_kwargs.update(kwargs)
        event_args.append(event_kwargs)
    if not returner_kwargs:
        returner_kwargs = {}
    opts = {
        'with_grains': with_grains,
        'with_pillar': with_pillar,
        'preload_grains': preload_grains,
        'preload_pillar': preload_pillar,
        'default_grains': default_grains,
        'default_pillar': default_pillar,
        'preload_targeting': preload_targeting,
        'invasive_targeting': invasive_targeting,
        'args': args,
        'cache_grains': cache_grains,
        'cache_pillar': cache_pillar,
        'use_cached_grains': use_cached_grains,
        'use_cached_pillar': use_cached_pillar,
        'use_existing_proxy': use_existing_proxy,
        'no_connect': no_connect,
        'test_ping': test_ping,
        'tgt': tgt,
        'tgt_type': tgt_type,
        'failhard': failhard,
        'timeout': timeout,
        'returner': returner,
        'returner_config': returner_config,
        'returner_kwargs': returner_kwargs,
    }
    opts.update(kwargs)
    if events:
        __salt__['event.send'](
            'proxy/runner/{jid}/new'.format(jid=jid),
            {
                'fun': salt_function,
                'minions': minions,
                'arg': event_args,
                'jid': jid,
                'tgt': tgt,
                'tgt_type': tgt_type,
                'user': __pub_user,
            },
        )
    if not existing_minions:
        existing_minions = []
    down_minions = []

    progress_bar = None
    if progress and HAS_PROGRESSBAR:
        progress_bar = progressbar.ProgressBar(
            max_value=len(minions), enable_colors=True, redirect_stdout=True
        )
    ret_queue = multiprocessing.Queue()
    done_queue = multiprocessing.Queue()
    if not static:
        thread = threading.Thread(
            target=_receive_replies_async, args=(ret_queue, done_queue, progress_bar)
        )
        thread.daemon = True
        thread.start()
    else:
        static_queue = multiprocessing.Queue()
        thread = threading.Thread(
            target=_receive_replies_sync,
            args=(ret_queue, static_queue, done_queue, progress_bar),
        )
        thread.daemon = True
        thread.start()

    ret = {}
    sproxy_minions = list(set(minions) - set(existing_minions))
    if batch_size:
        if '%' in str(batch_size):
            percent = int(batch_size.replace('%', ''))
            batch_size = len(minions) * percent / 100
        batch_size = int(batch_size)
        batch_count = int(len(minions) / batch_size) + (
            1 if len(minions) % batch_size else 0
        )
        existing_batch_size = int(
            math.ceil(len(existing_minions) * batch_size / float(len(minions)))
        )
        sproxy_batch_size = batch_size - existing_batch_size
    else:
        # when no explicit batch requested, we'll execute the command on the
        # existing minions without any batching (i.e., on all the matched
        # minions at once), while sproxy ones are executed in as many CPUs are
        # available.
        sproxy_batch_size = multiprocessing.cpu_count()
        existing_batch_size = len(existing_minions)
        batch_count = int(len(sproxy_minions) / sproxy_batch_size) + (
            1 if len(sproxy_minions) % sproxy_batch_size else 0
        )
    cli_batch = None
    if existing_batch_size > 0:
        # When there are existing Minions matching the target, use the native
        # batching function to execute against these Minions.
        log.debug('Executing against the existing Minions')
        log.debug(existing_minions)
        batch_opts = copy.deepcopy(__opts__)
        batch_opts['batch'] = str(existing_batch_size)
        batch_opts['tgt'] = existing_minions
        batch_opts['tgt_type'] = 'list'
        batch_opts['fun'] = salt_function
        batch_opts['arg'] = event_args
        batch_opts['batch_wait'] = batch_wait
        batch_opts['selected_target_option'] = 'list'
        batch_opts['return'] = returner
        batch_opts['ret_config'] = returner_config
        batch_opts['ret_kwargs'] = returner_kwargs
        if test_ping:
            cli_batch = Batch(batch_opts, quiet=True)
        else:
            cli_batch = NoPingBatch(batch_opts, quiet=True)
        log.debug('Batching detected the following Minions responsive')
        log.debug(cli_batch.minions)
        if cli_batch.down_minions:
            log.warning(
                'The following existing Minions connected to the Master '
                'seem to be unresponsive: %s',
                ', '.join(cli_batch.down_minions),
            )
            down_minions = cli_batch.down_minions
            for minion in down_minions:
                ret_queue.put(
                    (
                        {minion: 'Minion did not return. [Not connected]'},
                        salt.defaults.exitcodes.EX_UNAVAILABLE,
                    )
                )

    log.info(
        '%d devices matched the target, executing in %d batches',
        len(minions),
        batch_count,
    )
    batch_stop_queue = multiprocessing.Queue()
    sproxy_stop_queue = multiprocessing.Queue()
    # This dance with the batch_stop_queue and sproxy_stop_queue is necessary
    # in order to make sure the execution stops at the same time (either at the
    # very end, or when the iteration must be interrupted - e.g., due to
    # failhard condition).
    # sproxy_stop_queue signalises to the batch execution that the sproxy
    # sequence is over (not under normal circumstances, but interrupted forcibly
    # therefore it tells to the batch to stop immediately). In a similar way,
    # batch_stop_queue is required at the very end to make sure we're sending
    # the sentinel signaling at the very end for the display thread -- for
    # example there can be situations when the sproxy execution may be empty as
    # all the targets are existing proxies, so the display must wait.
    if cli_batch:
        existing_proxy_thread = threading.Thread(
            target=_existing_proxy_cli_batch,
            args=(cli_batch, ret_queue, batch_stop_queue, sproxy_stop_queue),
        )
        existing_proxy_thread.daemon = True
        existing_proxy_thread.start()
    else:
        # If there's no batch to execute (i.e., no existing devices to run
        # against), just need to signalise that there's no need to wait for this
        # one to complete.
        batch_stop_queue.put(0)

    log.debug(
        'Executing sproxy normal run on the following devices (%d batch size):',
        sproxy_batch_size,
    )
    log.debug(sproxy_minions)

    with multiprocessing.Manager() as manager:
        # Put the sproxy execution details into a Queue, from where the
        # processes from the bucket (see below) will pick them up whenever
        # there's room for another process to start up.
        sproxy_execute_queue = manager.Queue()
        for minion_id in sproxy_minions:
            device_opts = copy.deepcopy(opts)
            if roster_targets and isinstance(roster_targets, dict):
                device_opts['roster_opts'] = roster_targets.get(minion_id, {}).get(
                    'minion_opts'
                )
            sproxy_execute_queue.put((minion_id, device_opts))

        timeout_devices = manager.list()
        failed_devices = manager.list()
        unreachable_devices = manager.list()

        device_count = 0
        sproxy_processes = []
        stop_iteration = False

        # In the sequence below, we'll have a process bucket with a maximum size
        # which is the batch size, which will make room best efforts for
        # processes to be started up whenever there's a new process finishing
        # the task (or forcibly stopped due to timeout).
        while not sproxy_execute_queue.empty() and not stop_iteration:
            if len(sproxy_processes) >= sproxy_batch_size:
                # Wait for the bucket to make room for another process.
                time.sleep(0.001)
                continue
            minion_id, device_opts = sproxy_execute_queue.get()
            log.debug('Starting execution for %s', minion_id)
            device_proc = multiprocessing.Process(
                target=_salt_call_and_return,
                name=minion_id,
                args=(
                    minion_id,
                    salt_function,
                    ret_queue,
                    unreachable_devices,
                    failed_devices,
                    event_args,
                    jid,
                    events,
                ),
                kwargs=device_opts,
            )
            device_proc.start()
            sproxy_processes.append(device_proc)
            device_count += 0

            processes = sproxy_processes[:]
            for proc in processes:
                if failhard and proc.exitcode:
                    stop_iteration = True

                if not sproxy_execute_queue.empty() and len(processes) < min(
                    len(sproxy_minions), sproxy_batch_size
                ):
                    # Wait to fill up the sproxy processes bucket, and only then
                    # start evaluating.
                    # Why `min()`? It is possible that we can run on a smaller
                    # set of devices than the batch size.
                    continue

                # Wait `timeout` seconds for the processes to execute.
                proc.join(timeout=timeout)
                if proc.is_alive():
                    # If the process didn't finish the task, it means it's past
                    # the timeout value, time to kiss it goodbye.
                    log.info(
                        'Terminating the process for %s, as it didn\'t reply within %d seconds',
                        proc._name,
                        timeout,
                    )
                    sproxy_processes.remove(proc)
                    if not hide_timeout:
                        ret_queue.put(
                            (
                                {proc._name: 'Minion did not return. [No response]'},
                                salt.defaults.exitcodes.EX_UNAVAILABLE,
                            )
                        )
                    # return code EX_UNAVAILABLE on process timeout?
                    retcode = max(retcode, salt.defaults.exitcodes.EX_UNAVAILABLE)
                    timeout_devices.append(proc._name)

                if proc.exitcode and isinstance(proc.exitcode, int):
                    retcode = max(retcode, proc.exitcode)

                # Terminate the process, making room for a new one.
                proc.terminate()
                if proc in sproxy_processes:
                    # proc may no longer be in sproxy_processes, if it has been
                    # already removed in the section above when exiting the loop
                    # forcibly.
                    sproxy_processes.remove(proc)

            if stop_iteration:
                log.error('Exiting as an error has occurred')
                ret_queue.put((_SENTINEL, salt.defaults.exitcodes.EX_GENERIC))
                sproxy_stop_queue.put(_SENTINEL)
                for proc in sproxy_processes:
                    proc.terminate()
                raise StopIteration

            if len(processes) < min(len(sproxy_minions), sproxy_batch_size):
                continue

            if batch_wait:
                log.debug(
                    'Waiting %f seconds before executing the next batch', batch_wait
                )
                time.sleep(batch_wait)

        # Waiting for the existing proxy batch to finish.
        while batch_stop_queue.empty():
            time.sleep(0.001)
        batch_retcode = batch_stop_queue.get()
        retcode = max(retcode, batch_retcode)

        # Prepare to quit.
        ret_queue.put((_SENTINEL, 0))
        # Wait a little to dequeue and print before throwing the progressbar,
        # the summary, etc.
        while done_queue.empty():
            time.sleep(0.001)

        if progress_bar:
            progress_bar.finish()

        if static:
            resp = {}
            while True:
                ret, _retcode = static_queue.get()
                retcode = max(retcode, _retcode)
                if ret == _SENTINEL:
                    break
                resp.update(ret)

        if summary:
            salt.utils.stringutils.print_cli('\n')
            salt.utils.stringutils.print_cli(
                '-------------------------------------------'
            )
            salt.utils.stringutils.print_cli('Summary')
            salt.utils.stringutils.print_cli(
                '-------------------------------------------'
            )
            salt.utils.stringutils.print_cli(
                '# of devices targeted: {0}'.format(len(minions))
            )
            salt.utils.stringutils.print_cli(
                '# of devices returned: {0}'.format(
                    len(minions) - len(timeout_devices) - len(unreachable_devices)
                )
            )
            salt.utils.stringutils.print_cli(
                '# of devices that did not return: {0}'.format(len(timeout_devices))
            )
            salt.utils.stringutils.print_cli(
                '# of devices with errors: {0}'.format(len(failed_devices))
            )
            salt.utils.stringutils.print_cli(
                '# of devices unreachable: {0}'.format(len(unreachable_devices))
            )
            if verbose:
                if timeout_devices:
                    salt.utils.stringutils.print_cli(
                        (
                            '\nThe following devices didn\'t return (timeout):'
                            '\n - {0}'.format('\n - '.join(timeout_devices))
                        )
                    )
                if failed_devices:
                    salt.utils.stringutils.print_cli(
                        (
                            '\nThe following devices returned "bad" output:'
                            '\n - {0}'.format('\n - '.join(failed_devices))
                        )
                    )
                if unreachable_devices:
                    salt.utils.stringutils.print_cli(
                        (
                            '\nThe following devices are unreachable:'
                            '\n - {0}'.format('\n - '.join(unreachable_devices))
                        )
                    )
            salt.utils.stringutils.print_cli(
                '-------------------------------------------'
            )
            if events:
                __salt__['event.send'](
                    'proxy/runner/{jid}/summary'.format(jid=jid),
                    {
                        'tgt': tgt,
                        'tgt_type': tgt_type,
                        'fun': salt_function,
                        'fun_args': event_args,
                        'jid': jid,
                        'user': __pub_user,
                        'retcode': retcode,
                        'matched_minions': minions,
                        'existing_minions': existing_minions,
                        'sproxy_minions': sproxy_minions,
                        'timeout_minions': list(timeout_devices),
                        'down_minions': list(down_minions),
                        'unreachable_devices': list(unreachable_devices),
                        'failed_minions': list(failed_devices),
                    },
                )
    __context__['retcode'] = retcode
    if retcode != salt.defaults.exitcodes.EX_OK:
        salt.utils.stringutils.print_cli(
            'ERROR: Minions returned with non-zero exit code'
        )
    return resp


def execute(
    tgt,
    salt_function=None,
    tgt_type='glob',
    roster=None,
    preview_target=False,
    target_details=False,
    timeout=60,
    with_grains=True,
    with_pillar=True,
    preload_grains=True,
    preload_pillar=True,
    default_grains=None,
    default_pillar=None,
    args=(),
    batch_size=None,
    batch_wait=0,
    static=False,
    events=True,
    cache_grains=True,
    cache_pillar=True,
    use_cached_grains=True,
    use_cached_pillar=True,
    use_existing_proxy=False,
    no_connect=False,
    test_ping=False,
    target_cache=False,
    target_cache_timeout=60,
    preload_targeting=False,
    invasive_targeting=False,
    failhard=False,
    summary=False,
    verbose=False,
    show_jid=False,
    progress=False,
    hide_timeout=False,
    sync_roster=False,
    sync_modules=False,
    sync_grains=False,
    sync_all=False,
    returner='',
    returner_config='',
    returner_kwargs=None,
    **kwargs
):
    '''
    Invoke a Salt function on the list of devices matched by the Roster
    subsystem.

    tgt
        The target expression, e.g., ``*`` for all devices, or ``host1,host2``
        for a list, etc. The ``tgt_list`` argument must be used accordingly,
        depending on the type of this expression.

    salt_function
        The name of the Salt function to invoke.

    tgt_type: ``glob``
        The type of the ``tgt`` expression. Choose between: ``glob`` (default),
        ``list``, ``pcre``, ``rage``, or ``nodegroup``.

    roster: ``None``
        The name of the Roster to generate the targets. Alternatively, you can
        specify the name of the Roster by configuring the ``proxy_roster``
        option into the Master config.

    preview_target: ``False``
        Return the list of Roster targets matched by the ``tgt`` and
        ``tgt_type`` arguments.

    preload_grains: ``True``
        Whether to preload the Grains before establishing the connection with
        the remote network device.

    default_grains:
        Dictionary of the default Grains to make available within the functions
        loaded.

    with_grains: ``True``
        Whether to load the Grains modules and collect Grains data and make it
        available inside the Execution Functions.
        The Grains will be loaded after opening the connection with the remote
        network device.

    default_pillar:
        Dictionary of the default Pillar data to make it available within the
        functions loaded.

    with_pillar: ``True``
        Whether to load the Pillar modules and compile Pillar data and make it
        available inside the Execution Functions.

    arg
        The list of arguments to send to the Salt function.

    kwargs
        Key-value arguments to send to the Salt function.

    batch_size: None
        The size of each batch to execute.

    static: ``False``
        Whether to return the results synchronously (or return them as soon
        as the device replies).

    events: ``True``
        Whether should push events on the Salt bus, similar to when executing
        equivalent through the ``salt`` command.

    use_cached_pillar: ``True``
        Use cached Pillars whenever possible. If unable to gather cached data,
        it falls back to compiling the Pillar.

    use_cached_grains: ``True``
        Use cached Grains whenever possible. If unable to gather cached data,
        it falls back to collecting Grains.

    cache_pillar: ``True``
        Cache the compiled Pillar data before returning.

    cache_grains: ``True``
        Cache the collected Grains before returning.

    use_existing_proxy: ``False``
        Use the existing Proxy Minions when they are available (say on an
        already running Master).

    no_connect: ``False``
        Don't attempt to initiate the connection with the remote device.
        Default: ``False`` (it will initiate the connection).

    test_ping: ``False``
        When using the existing Proxy Minion with the ``use_existing_proxy``
        option, can use this argument to verify also if the Minion is
        responsive.

    target_cache: ``True``
        Whether to use the cached target matching results.

    target_cache_timeout: 60
        The duration to cache the target results for (in seconds).

    CLI Example:

    .. code-block:: bash

        salt-run proxy.execute_roster edge* test.ping
        salt-run proxy.execute_roster junos-edges test.ping tgt_type=nodegroup
    '''
    targets = []
    rtargets = None
    roster = roster or __opts__.get('proxy_roster', __opts__.get('roster'))

    saltenv = __opts__.get('saltenv', 'base')
    if sync_roster and not sync_all:
        __salt__['saltutil.sync_roster'](saltenv=saltenv)
    if sync_modules and not sync_all:
        __salt__['saltutil.sync_modules'](saltenv=saltenv)
    if sync_all:
        __salt__['saltutil.sync_all'](saltenv=saltenv)

    if not timeout:
        log.warning('Timeout set as 0, will wait for the devices to reply indefinitely')
        # Setting the timeout as None, because that's the value we need to pass
        # to multiprocessing's join() method to wait for the devices to reply
        # indefinitely.
        timeout = None

    if tgt_type == 'pillar_target':
        # When using the -I option on the CLI, the tgt_type passed on is called
        # `pillar_target`:
        # https://github.com/saltstack/salt/blob/e9e48b7fb6a688f4f22d74a849d58c1c156563d1/salt/utils/parsers.py#L1266
        # While if we want to use this against existing Minions, the option
        # needs to be just `pillar`:
        # https://github.com/saltstack/salt/blob/99385b50718d70d93fd5b83e61c0f4b3a402490c/salt/utils/minions.py#L359
        tgt_type = 'pillar'

    if preload_targeting or invasive_targeting:
        _tgt = '*'
        _tgt_type = 'glob'
    else:
        _tgt = tgt
        _tgt_type = tgt_type

    existing_minions = []
    if not roster or roster == 'None':
        log.info(
            'No Roster specified. Please use the ``roster`` argument, or set the ``proxy_roster`` option in the '
            'Master configuration.'
        )
        targets = []
        if use_existing_proxy:
            # When targeting exiting Proxies, we're going to look and match the
            # accepted keys
            log.debug('Requested to match the target based on the existing Minions')
            target_util = salt.utils.master.MasterPillarUtil(
                tgt,
                tgt_type,
                use_cached_grains=True,
                grains_fallback=False,
                opts=__opts__,
            )
            targets = target_util._tgt_to_list()
            existing_minions = targets[:]
        else:
            # Try a fuzzy match based on the exact target the user requested
            # only when not attempting to match an existing Proxy. If you do
            # want however, it won't be of much use, as the command is going to
            # be spread out to non-existing minions, so better turn off that
            # feature.
            log.debug('Trying a fuzzy match on the target')
            if tgt_type == 'list':
                targets = tgt[:]
            elif tgt_type == 'glob' and tgt != '*':
                targets = [tgt]
    else:
        targets = None
        if target_cache and not (invasive_targeting or preload_targeting):
            cache_bank = salt.cache.factory(__opts__)
            cache_key = hashlib.sha1(
                '{tgt}_{tgt_type}'.format(tgt=tgt, tgt_type=tgt_type).encode()
            ).hexdigest()
            cache_time_key = '{}_time'.format(cache_key)
            cache_time = cache_bank.fetch('_salt_sproxy_target', cache_time_key)
            if cache_time and time.time() - cache_time <= target_cache_timeout:
                log.debug('Loading the targets from the cache')
                targets = cache_bank.fetch('_salt_sproxy_target', cache_key)
        if not targets:
            rtargets = {}
            if use_existing_proxy:
                log.debug('Gathering the cached Grains from the existing Minions')
                cached_grains = __salt__['cache.grains'](tgt=tgt, tgt_type=tgt_type)
                for target, target_grains in cached_grains.items():
                    rtargets[target] = {'minion_opts': {'grains': target_grains}}
                    existing_minions.append(target)
            log.debug('Computing the target using the %s Roster', roster)
            __opts__['use_cached_grains'] = use_cached_grains
            __opts__['use_cached_pillar'] = use_cached_pillar
            roster_modules = salt.loader.roster(
                __opts__, runner=__salt__, whitelist=[roster]
            )
            if '.targets' not in roster:
                roster = '{mod}.targets'.format(mod=roster)
            rtargets_roster = roster_modules[roster](_tgt, tgt_type=_tgt_type)
            rtargets = salt.utils.dictupdate.merge(rtargets, rtargets_roster)
            targets = list(rtargets.keys())
            if target_cache and not (invasive_targeting or preload_targeting):
                cache_bank.store('_salt_sproxy_target', cache_key, targets)
                cache_bank.store('_salt_sproxy_target', cache_time_key, time.time())
    if preload_targeting or invasive_targeting:
        log.debug(
            'Loaded everything from the Roster, to start collecting Grains and Pillars:'
        )
    else:
        log.debug(
            'The target expression "%s" (%s) matched the following:', str(tgt), tgt_type
        )
    log.debug(targets)
    if not targets:
        return 'No devices matched your target. Please review your tgt / tgt_type arguments, or the Roster data source'
    if preview_target:
        return targets
    elif not salt_function:
        return 'Please specify a Salt function to execute.'
    jid = kwargs.get('__pub_jid')
    if not jid:
        if salt.version.__version_info__ >= (2018, 3, 0):
            jid = salt.utils.jid.gen_jid(__opts__)
        else:
            jid = salt.utils.jid.gen_jid()  # pylint: disable=no-value-for-parameter
    if verbose or show_jid:
        salt.utils.stringutils.print_cli('Executing job with jid {0}'.format(jid))
        salt.utils.stringutils.print_cli(
            '-------------------------------------------\n'
        )
    if events:
        __salt__['event.send'](jid, {'minions': targets})
    return execute_devices(
        targets,
        salt_function,
        tgt=tgt,
        tgt_type=tgt_type,
        with_grains=with_grains,
        preload_grains=preload_grains,
        with_pillar=with_pillar,
        preload_pillar=preload_pillar,
        default_grains=default_grains,
        default_pillar=default_pillar,
        args=args,
        batch_size=batch_size,
        batch_wait=batch_wait,
        static=static,
        events=events,
        cache_grains=cache_grains,
        cache_pillar=cache_pillar,
        use_cached_grains=use_cached_grains,
        use_cached_pillar=use_cached_pillar,
        use_existing_proxy=use_existing_proxy,
        existing_minions=existing_minions,
        no_connect=no_connect,
        roster_targets=rtargets,
        test_ping=test_ping,
        preload_targeting=preload_targeting,
        invasive_targeting=invasive_targeting,
        failhard=failhard,
        timeout=timeout,
        summary=summary,
        verbose=verbose,
        progress=progress,
        hide_timeout=hide_timeout,
        returner=returner,
        returner_config=returner_config,
        returner_kwargs=returner_kwargs,
        **kwargs
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution('salt_sproxy').version
except pkg_resources.DistributionNotFound:
    __version__ = 'Not installed'

__version_info__ = tuple(__version__.split('.'))

all = ['__version__', '__version_info__']



# -*- coding: utf-8 -*-
'''
Load the list of devices from the Pillar.

Simply configure the ``roster`` option to point to this module, while making
sure that the data is available. As the Pillar is data associated with a
specific Minion ID, you may need to ensure that the Pillar is correctly
associated with the Minion configured (default ``*``), under the exact key
required (default ``devices``). To adjust these options, you can provide the
following under the ``roster_pillar`` option in the Master configuration:

minion_id: ``*``
    The ID of the Minion to compile the data for. Default: ``*`` (any Minion).

pillar_key: ``devices``
    The Pillar field to pull the list of devices from. Default: ``devices``.

saltenv: ``base``
    The Salt environment to use when compiling the Pillar data.

pillarenv
    The Pillar environment to use when compiling the Pillar data.

Configuration example:

.. code-block:: yaml

    roster: pillar
    roster_pillar:
      minion_id: sproxy
      pillar_key: minions

With the following configuration, when executing
``salt-run pillar.show_pillar sproxy`` you should have under ``minions`` the
list of devices / Minions you want to manage.

.. hint::

    The Pillar data can either be provided as files, or using one or more
    External Pillars. Check out
    https://docs.saltstack.com/en/latest/ref/pillar/all/index.html
    for the complete list of available Pillar modules you can use.
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging

import salt_sproxy._roster

__virtualname__ = 'pillar'

log = logging.getLogger(__name__)


def targets(tgt, tgt_type='glob', **kwargs):
    '''
    Return the targets from External Pillar requested.
    '''
    roster_opts = __opts__.get('roster_pillar', {})
    minion_id = roster_opts.get('minion_id', kwargs.get('minion_id', '*'))
    pillar_key = roster_opts.get('pillar_key', kwargs.get('pillar_key', 'devices'))
    saltenv = roster_opts.get('saltenv', kwargs.get('saltenv', 'base'))
    pillarenv = roster_opts.get('pillarenv', kwargs.get('pillarenv'))
    pillar = __runner__['pillar.show_pillar'](
        minion=minion_id, saltenv=saltenv, pillarenv=pillarenv
    )
    pillar_devices = pillar[pillar_key]
    log.debug('Compiled the following list of devices from the Pillar')
    log.debug(pillar_devices)
    pool = {
        device.pop('id', device.pop('name')): {'minion_opts': device}
        for device in pillar_devices
    }
    pool = salt_sproxy._roster.load_cache(
        pool, __runner__, __opts__, tgt, tgt_type=tgt_type
    )
    engine = salt_sproxy._roster.TGT_FUN[tgt_type]
    return engine(pool, tgt, opts=__opts__)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import sys
import optparse
import multiprocessing

import salt_sproxy.version

from salt.ext import six
import salt.version
import salt.utils.args
import salt.utils.parsers
import salt.config as config
from salt.ext.six.moves import map
from salt.ext.six.moves import range

try:
    from jnpr.junos import __version__ as jnpr_version

    # Ain't Juniper awkward?
except ImportError:
    jnpr_version = None

CPU_COUNT = multiprocessing.cpu_count()


def salt_information():
    '''
    Return version of Salt and salt-sproxy.
    '''
    yield 'Salt', salt.version.__version__
    yield 'Salt SProxy', salt_sproxy.version.__version__


def dependency_information(include_salt_cloud=False):
    '''
    Report versions of library dependencies.

    This function has been ported from
    https://github.com/saltstack/salt/blob/develop/salt/version.py
    and extended here to collect the version information for several more
    libraries that may be necessary for various Proxy (or Execution) Modules.
    '''
    libs = [
        ('Python', None, sys.version.rsplit('\n')[0].strip()),
        ('NAPALM', 'napalm', '__version__'),
        ('Netmiko', 'netmiko', '__version__'),
        ('junos-eznc', None, jnpr_version),
        ('ncclient', 'ncclient', '__version__'),
        ('paramiko', 'paramiko', '__version__'),
        ('pyeapi', 'pyeapi', '__version__'),
        ('textfsm', 'textfsm', '__version__'),
        ('jxmlease', 'jxmlease', '__version__'),
        ('scp', 'scp', '__version__'),
        ('PyNSO', 'pynso', '__version__'),
        ('Ansible', 'ansible', '__version__'),
        ('PyNetBox', 'pynetbox', '__version__'),
        ('Jinja2', 'jinja2', '__version__'),
        ('M2Crypto', 'M2Crypto', 'version'),
        ('msgpack-python', 'msgpack', 'version'),
        ('msgpack-pure', 'msgpack_pure', 'version'),
        ('pycrypto', 'Crypto', '__version__'),
        ('pycryptodome', 'Cryptodome', 'version_info'),
        ('PyYAML', 'yaml', '__version__'),
        ('PyZMQ', 'zmq', '__version__'),
        ('ZMQ', 'zmq', 'zmq_version'),
        ('Mako', 'mako', '__version__'),
        ('Tornado', 'tornado', 'version'),
        ('timelib', 'timelib', 'version'),
        ('dateutil', 'dateutil', '__version__'),
        ('pygit2', 'pygit2', '__version__'),
        ('libgit2', 'pygit2', 'LIBGIT2_VERSION'),
        ('smmap', 'smmap', '__version__'),
        ('cffi', 'cffi', '__version__'),
        ('pycparser', 'pycparser', '__version__'),
        ('gitdb', 'gitdb', '__version__'),
        ('gitpython', 'git', '__version__'),
        ('python-gnupg', 'gnupg', '__version__'),
        ('docker-py', 'docker', '__version__'),
    ]

    if include_salt_cloud:
        libs.append(('Apache Libcloud', 'libcloud', '__version__'))

    for name, imp, attr in libs:
        if imp is None:
            yield name, attr
            continue
        try:
            imp = __import__(imp)
            version = getattr(imp, attr)
            if callable(version):
                version = version()
            if isinstance(version, (tuple, list)):
                version = '.'.join(map(str, version))
            yield name, version
        except Exception:
            yield name, None


salt.version.salt_information = salt_information
salt.version.dependency_information = dependency_information


class SaltStandaloneProxyOptionParser(
    six.with_metaclass(
        salt.utils.parsers.OptionParserMeta,
        salt.utils.parsers.OptionParser,
        salt.utils.parsers.ConfigDirMixIn,
        salt.utils.parsers.MergeConfigMixIn,
        salt.utils.parsers.TimeoutMixIn,
        salt.utils.parsers.LogLevelMixIn,
        salt.utils.parsers.HardCrashMixin,
        salt.utils.parsers.SaltfileMixIn,
        salt.utils.parsers.ExtendedTargetOptionsMixIn,
        salt.utils.parsers.OutputOptionsMixIn,
        salt.utils.parsers.ArgsStdinMixIn,
        salt.utils.parsers.ProfilingPMixIn,
        salt.utils.parsers.EAuthMixIn,
        salt.utils.parsers.NoParseMixin,
    )
):

    default_timeout = 60

    description = (
        r'''
  ___          _   _       ___   ___
 / __|  __ _  | | | |_    / __| | _ \  _ _   ___  __ __  _  _
 \__ \ / _` | | | |  _|   \__ \ |  _/ | '_| / _ \ \ \ / | || |
 |___/ \__,_| |_|  \__|   |___/ |_|   |_|   \___/ /_\_\  \_, |
                                                         |__/

'''
        'salt-sproxy is a tool to invoke arbitrary Salt functions on a group\n'
        'of (network) devices connecting through a Salt Proxy Minion, without\n'
        'having the Proxy Minion services up and running (or the Master).\n'
    )
    VERSION = salt_sproxy.version.__version__

    usage = '%prog [options] <target> <function> [arguments]'

    epilog = (
        'You can find additional help about %prog at '
        'https://salt-sproxy.readthedocs.io/en/latest/'
    )

    # ConfigDirMixIn config filename attribute
    _config_filename_ = 'master'

    # LogLevelMixIn attributes
    _default_logging_level_ = config.DEFAULT_MASTER_OPTS['log_level']
    _default_logging_logfile_ = config.DEFAULT_MASTER_OPTS['log_file']

    def format_description(self, formatter):
        return self.description

    def _mixin_setup(self):
        self.add_option(
            '-r', '--roster', default=False, help='The name of the Salt Roster to use.'
        )
        self.add_option(
            '--roster-file',
            dest='roster_file',
            help='Absolute path to the Roster file to use.',
        )
        self.add_option(
            '-s',
            '--static',
            default=False,
            action='store_true',
            help=('Return the data from devices as a group after they all return.'),
        )
        self.add_option(
            "--async",
            default=False,
            dest="async",
            action="store_true",
            help=('Run the salt-sproxy command but don\'t wait for a reply.'),
        )
        self.add_option(
            '--dont-cache-grains',
            default=False,
            action='store_true',
            help=('Do not cache the collected Grains for the sproxy devices.'),
        )
        self.add_option(
            '--dont-cache-pillar',
            default=False,
            action='store_true',
            help=('Do not cache the compiled Pillar for the sproxy devices.'),
        )
        self.add_option(
            '--no-cached-grains',
            default=False,
            action='store_true',
            help='Do not use the available cached Grains (if any).',
        )
        self.add_option(
            '--no-cached-pillar',
            default=False,
            action='store_true',
            help='Do not use the available cached Pillar (if any)',
        )
        self.add_option(
            '--no-grains',
            default=False,
            action='store_true',
            help=(
                'Do not attempt to collect Grains at all. Use with care, it '
                'may lead to unexpected results.'
            ),
        )
        self.add_option(
            '-i',
            '--ignore-host-keys',
            default=False,
            action='store_true',
            dest='ignore_host_keys',
            help=(
                'By default ssh host keys are honored and connections will ask '
                'for approval. Use this option to disable StrictHostKeyChecking.'
            ),
        )
        self.add_option(
            '--no-host-keys',
            default=False,
            action='store_true',
            dest='no_host_keys',
            help=(
                'Fully ignores ssh host keys which by default are honored and '
                'connections would ask for approval. Useful if the host key of '
                'a remote server has changed and would still error with '
                '--ignore-host-keys.'
            ),
        )
        self.add_option(
            '--identities-only',
            default=False,
            action='store_true',
            dest='identities_only',
            help=(
                'Use the only authentication identity files configured in the '
                'ssh_config files. See ``IdentitiesOnly`` flag in man ssh_config.'
            ),
        )
        self.add_option(
            '--priv',
            dest='priv',
            help=('Specify the SSH private key file to be used for authentication.'),
        )
        self.add_option(
            '--priv-passwd',
            dest='priv_passwd',
            help=('Specify the SSH private key file\'s passphrase when required.'),
        )
        self.add_option(
            '--preload-targeting',
            default=False,
            action='store_true',
            help=(
                'Preload Grains for all the devices before targeting.'
                'This is useful to match the devices on dynamic Grains that '
                'do not require the connection with the remote device - e.g., '
                'Grains collected from an external API, etc.'
            ),
        )
        self.add_option(
            '--invasive-targeting',
            default=False,
            action='store_true',
            help=(
                'Collect all the possible data from every device salt-sproxy '
                'is aware of, before distributing the commands. '
                'In other words, this option tells salt-sproxy to connect to '
                'every possible device defined in the Roster of choice, collect '
                'Grains, compile Pillars, etc., and only then execute the '
                'command against the devices matching the target expression.'
                'Use with care, as this may significantly reduce the '
                'performances, but this is the price paid to be able to target '
                'using device properties. '
                'Consider using this option in conjunction with --cache-grains '
                'and / or --cache-pillar to cache the Grains and the Pillars to '
                're-use them straight away next time.'
            ),
        )
        self.add_option(
            '--no-pillar',
            default=False,
            action='store_true',
            help=(
                'Do not compile Pillar at all. Use with care, it may lead to '
                'unexpected results.'
            ),
        )
        self.add_option(
            '-b',
            '--batch',
            '--batch-size',
            dest='batch_size',
            help=(
                'The number of devices to connect to in parallel. '
                'Default: {} (number of CPUs on your machine)'.format(CPU_COUNT)
            ),
        )
        self.add_option(
            '--preview-target',
            dest='preview_target',
            action='store_true',
            help='Show the devices expected to match the target.',
        )
        self.add_option(
            '--sync-all',
            dest='sync_all',
            action='store_true',
            help=(
                'Load the all extension modules provided with salt-sproxy, as '
                'well as the extension modules from your own environment.'
            ),
        )
        self.add_option(
            '--sync-grains',
            dest='sync_grains',
            action='store_true',
            help=(
                'Re-sync the Grains modules. Useful if you have custom Grains '
                'modules in your own environment.'
            ),
        )
        self.add_option(
            '--sync-modules',
            dest='sync_modules',
            action='store_true',
            help=('Load the salt-sproxy Execution modules.'),
        )
        self.add_option(
            '--sync-roster',
            dest='sync_roster',
            action='store_true',
            help=(
                'Synchronise the Roster modules (both salt-sproxy native '
                'and provided by the user in their own environment).'
            ),
        )
        self.add_option(
            '--sync-proxy',
            dest='sync_proxy',
            action='store_true',
            help=('Load the salt-sproxy Proxy modules.'),
        )
        self.add_option(
            '--sync-executors',
            dest='sync_executors',
            action='store_true',
            help=('Load the salt-sproxy Executor modules.'),
        )
        self.add_option(
            '--saltenv',
            dest='saltenv_cli',
            help='The Salt environment name to load module and files from',
        )
        self.add_option(
            '--events',
            dest='events',
            action='store_true',
            help=(
                'Whether should put the events on the Salt bus (mostly '
                'useful when having a Master running).'
            ),
        )
        self.add_option(
            '--use-proxy',
            '--use-existing-proxy',
            '--use-existing-minion',
            dest='use_existing_proxy',
            action='store_true',
            help=(
                'Use the existing Proxy Minions to execute the commands, '
                'whenever available.'
            ),
        )
        self.add_option(
            '--pillar-root',
            default=None,
            help='Set this directory as the base pillar root.',
        )
        self.add_option(
            '--file-root',
            default=None,
            help='Set this directory as the base file root.',
        )
        self.add_option(
            '--states-dir',
            default=None,
            help='Set this directory to search for additional states.',
        )
        self.add_option(
            '-m',
            '--module-dirs',
            dest='module_dirs_cli',
            default=[],
            action='append',
            help=(
                'Specify an additional directory to pull modules from. '
                'Multiple directories can be provided by passing '
                '`-m/--module-dirs` multiple times.'
            ),
        )
        self.add_option(
            '--installation-path',
            dest='installation_path',
            action='store_true',
            help=('Display the absolute path to where salt-sproxy is installed.'),
        )
        self.add_option(
            '--display-file-roots',
            dest='display_file_roots',
            action='store_true',
            help=(
                'Display the file_roots option you would need to configure '
                'in order to use the salt-sproxy extension modules directly, '
                'and, implicitly, leverage the event-driven methodology and '
                'the Salt REST API.'
            ),
        )
        self.add_option(
            '--save-file-roots',
            dest='save_file_roots',
            action='store_true',
            help=(
                'Saves the file_roots configuration so you can start '
                'leveraging the event-driven automation and the Salt REST API.'
            ),
        )
        self.add_option(
            '--config-dump',
            dest='config_dump',
            default=False,
            action='store_true',
            help='Dump the salt-sproxy configuration values',
        )
        self.add_option(
            '--no-connect',
            dest='no_connect',
            action='store_true',
            default=False,
            help=(
                'Do not initiate the connection with the device, only use '
                'cached data to compile data and execute Salt functions that '
                'do not require the actual connection with the device.'
            ),
        )
        self.add_option(
            '--test-ping',
            dest='test_ping',
            action='store_true',
            default=False,
            help=(
                'When using together with --use-existing-proxy, this option can'
                ' help to ensure the existing Proxy Minion is responsive (not '
                'only up and running, by executing a ping test.'
            ),
        )
        self.add_option(
            '--failhard',
            dest='failhard',
            action='store_true',
            default=False,
            help='Stop execution at the first execution error.',
        )
        self.add_option(
            '--target-cache',
            dest='target_cache',
            action='store_true',
            default=False,
            help=('Cache the list of devices matched by your target expression.'),
        )
        group = self.output_options_group = optparse.OptionGroup(
            self, 'Output Options', 'Configure your preferred output format.'
        )
        self.add_option_group(group)

        group.add_option(
            '-q',
            '--quiet',
            default=False,
            action='store_true',
            help='Do not display the results of the run.',
        )
        self.add_option(
            '--summary',
            default=False,
            action='store_true',
            help='Display salt execution summary information.',
        )
        self.add_option(
            '-v',
            '--verbose',
            default=False,
            action='store_true',
            help='Turn on command verbosity, display jid and detailed summary.',
        )
        self.add_option(
            '--show-jid',
            default=False,
            action='store_true',
            help='Display jid without the additional output of --verbose.',
        )
        self.add_option(
            '--hide-timeout',
            default=False,
            action='store_true',
            help='Hide devices that timeout.',
        )
        self.add_option(
            '--batch-wait',
            default=0,
            type=float,
            help=(
                'Wait the specified time in seconds after each batch is done'
                'before executing the next one.'
            ),
        )
        self.add_option(
            '-p',
            '--progress',
            default=False,
            action='store_true',
            help='Display a progress graph.',
        )
        self.add_option(
            '--return',
            dest='returner',
            default='',
            metavar='RETURNER',
            help=(
                'The name of the Returner module to use for sending data to '
                'various external systems.'
            ),
        )
        self.add_option(
            '--return-config',
            dest='returner_config',
            default='',
            metavar='RETURNER_CONF',
            help='Specify an alternative Returner config.',
        )
        self.add_option(
            '--return-kwargs',
            dest='returner_kwargs',
            default={},
            metavar='RETURNER_KWARGS',
            help='Set Returner options at the command line.',
        )
        self.add_option(
            "-d",
            "--doc",
            "--documentation",
            dest="doc",
            default=False,
            action="store_true",
            help=(
                'Return the documentation for the specified module or for '
                'all modules if none are specified.'
            ),
        )

    # Everything else that follows here is verbatim copy from
    # https://github.com/saltstack/salt/blob/develop/salt/utils/parsers.py
    def _mixin_after_parsed(self):
        if (
            self.options.display_file_roots
            or self.options.installation_path
            or self.options.save_file_roots
            or self.options.config_dump
        ):
            # Insert dummy arg when displaying the file_roots
            self.args.append('not_a_valid_target')
            self.args.append('not_a_valid_command')
        elif self.options.doc:
            if len(self.args) == 1:
                self.args.insert(0, 'not_a_valid_target')
            elif len(self.args) == 0:
                self.args.append('not_a_valid_target')
                self.args.append('*')

        if self.options.list:
            try:
                if ',' in self.args[0]:
                    self.config['tgt'] = self.args[0].replace(' ', '').split(',')
                else:
                    self.config['tgt'] = self.args[0].split()
            except IndexError:
                self.exit(42, '\nCannot execute command without defining a target.\n\n')
        else:
            try:
                self.config['tgt'] = self.args[0]
            except IndexError:
                self.exit(42, '\nCannot execute command without defining a target.\n\n')

        if self.options.preview_target:
            # Insert dummy arg which won't be used
            self.args.append('not_a_valid_command')

        # Detect compound command and set up the data for it
        if self.args:
            try:
                if ',' in self.args[1]:
                    self.config['fun'] = self.args[1].split(',')
                    self.config['arg'] = [[]]
                    cmd_index = 0
                    if (
                        self.args[2:].count(self.options.args_separator)
                        == len(self.config['fun']) - 1
                    ):
                        # new style parsing: standalone argument separator
                        for arg in self.args[2:]:
                            if arg == self.options.args_separator:
                                cmd_index += 1
                                self.config['arg'].append([])
                            else:
                                self.config['arg'][cmd_index].append(arg)
                    else:
                        # old style parsing: argument separator can be inside args
                        for arg in self.args[2:]:
                            if self.options.args_separator in arg:
                                sub_args = arg.split(self.options.args_separator)
                                for sub_arg_index, sub_arg in enumerate(sub_args):
                                    if sub_arg:
                                        self.config['arg'][cmd_index].append(sub_arg)
                                    if sub_arg_index != len(sub_args) - 1:
                                        cmd_index += 1
                                        self.config['arg'].append([])
                            else:
                                self.config['arg'][cmd_index].append(arg)
                        if len(self.config['fun']) > len(self.config['arg']):
                            self.exit(
                                42,
                                'Cannot execute compound command without '
                                'defining all arguments.\n',
                            )
                        elif len(self.config['fun']) < len(self.config['arg']):
                            self.exit(
                                42,
                                'Cannot execute compound command with more '
                                'arguments than commands.\n',
                            )
                    # parse the args and kwargs before sending to the publish
                    # interface
                    for i in range(len(self.config['arg'])):
                        self.config['arg'][i] = salt.utils.args.parse_input(
                            self.config['arg'][i], no_parse=self.options.no_parse
                        )
                else:
                    self.config['fun'] = self.args[1]
                    self.config['arg'] = self.args[2:]
                    # parse the args and kwargs before sending to the publish
                    # interface
                    self.config['arg'] = salt.utils.args.parse_input(
                        self.config['arg'], no_parse=self.options.no_parse
                    )
            except IndexError:
                self.exit(42, '\nIncomplete options passed.\n\n')

    def setup_config(self):
        defaults = config.DEFAULT_MASTER_OPTS.copy()
        defaults['timeout'] = 60
        return config.client_config(self.get_config_file_path(), defaults=defaults)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import sys
import inspect
import logging

import salt.netapi
import salt.scripts
from salt.ext import six
import salt.utils.parsers
from salt.scripts import _install_signal_handlers
from salt_sproxy._runners.proxy import execute as sproxy_execute

log = logging.getLogger(__name__)


def _prep_kwargs(kwargs, opts):
    '''
    Gather the sproxy execute argument from the Master opts when available.
    '''
    execute_args = set(inspect.getfullargspec(sproxy_execute)[0])
    sapi_args = execute_args - {'tgt', 'salt_function', 'tgt_type', 'static', 'timeout'}
    for arg in sapi_args:
        if arg not in kwargs and arg in opts:
            kwargs[arg] = opts[arg]
    return kwargs


def sapi_sproxy(
    self, tgt, fun, tgt_type='glob', timeout=None, full_return=False, **kwargs
):
    '''
    Shortcut to invoke an arbitrary Salt function via sproxy.
    '''
    kwargs.update(
        {'salt_function': fun, 'tgt': tgt, 'tgt_type': tgt_type, 'static': True}
    )
    kwargs = _prep_kwargs(kwargs, self.opts)
    log.debug('New kwargs:')
    log.debug(kwargs)
    return salt.netapi.NetapiClient.runner(
        self, 'proxy.execute', timeout=timeout, full_return=full_return, **kwargs
    )


def sapi_sproxy_async(
    self, tgt, fun, tgt_type='glob', timeout=None, full_return=False, **kwargs
):
    '''
    Shortcut to invoke an arbitrary Salt function via sproxy, asynchronously.
    '''
    kwargs.update(
        {'salt_function': fun, 'tgt': tgt, 'tgt_type': tgt_type, 'static': True}
    )
    kwargs = _prep_kwargs(kwargs, self.opts)
    log.debug('New kwargs:')
    log.debug(kwargs)
    return salt.netapi.NetapiClient.runner_async(
        self, 'proxy.execute', timeout=timeout, full_return=full_return, **kwargs
    )


salt.netapi.NetapiClient.sproxy = sapi_sproxy
salt.netapi.NetapiClient.sproxy_async = sapi_sproxy_async
salt.netapi.CLIENTS = [
    name
    for name, _ in inspect.getmembers(
        salt.netapi.NetapiClient, predicate=inspect.ismethod if six.PY2 else None
    )
    if not (name == 'run' or name.startswith('_'))
]
salt.utils.parsers.SaltAPIParser.description = (
    'salt-sapi is an enhanced Salt API system that provides additional '
    'sproxy and sproxy_async clients, to simplify the usage of salt-sproxy '
    'through the Salt REST API'
)
salt.utils.parsers.SaltAPIParser.epilog = (
    'You can find additional help about %prog issuing "man %prog" '
    'or on https://salt-sproxy.readthedocs.io/ and '
    'https://docs.saltstack.com/en/latest/ref/cli/salt-api.html.'
)


def salt_sapi():
    '''
    The main function for salt-sapi.
    '''
    salt.scripts.salt_api()


def salt_sproxy():
    '''
    Execute a salt convenience routine.
    '''
    import salt_sproxy.cli

    if '' in sys.path:
        sys.path.remove('')
    client = salt_sproxy.cli.SaltStandaloneProxy()
    _install_signal_handlers(client)
    client.run()


if __name__ == '__main__':
    salt_sproxy()



# -*- coding: utf-8 -*-
# Copyright 2019-2020 Mircea Ulinic. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
'''
The CLI entry point module.
'''
from __future__ import absolute_import, print_function, unicode_literals

from salt_sproxy.parsers import SaltStandaloneProxyOptionParser

import os
import sys
import logging

import salt.runner
import salt.utils.parsers
from salt.utils.verify import check_user, verify_log
from salt.exceptions import SaltClientError
from salt.ext import six
import salt.defaults.exitcodes  # pylint: disable=W0611
import salt.utils.stringutils

try:
    from salt.utils.files import fopen
    from salt.utils.yamldumper import safe_dump
    from salt.utils.yamlloader import safe_load
    from salt.utils.profile import output_profile
    from salt.utils.profile import activate_profile
except ImportError:
    from salt.utils import fopen
    from salt.utils.yamldumper import safe_dump
    from salt.utils.yamlloader import load as safe_load
    from salt.utils import output_profile
    from salt.utils import activate_profile

log = logging.getLogger(__name__)


class SaltStandaloneProxy(SaltStandaloneProxyOptionParser):
    '''
    Used to execute Salt functions on a number of devices.
    '''

    def run(self):
        '''
        Execute salt-run
        '''
        self.parse_args()

        if self.config.get('config_dump'):
            sys.stdout.write(safe_dump(self.config, default_flow_style=False))
            return self.config

        # Setup file logging!
        self.setup_logfile_logger()
        verify_log(self.config)
        profiling_enabled = self.options.profiling_enabled
        curpath = os.path.dirname(os.path.realpath(__file__))
        saltenv = self.config.get('saltenv_cli', self.config.get('saltenv'))
        if not saltenv:
            saltenv = 'base'
        self.config['saltenv'] = saltenv
        if self.config.get('pillar_root'):
            log.info(
                'Setting and using %s as the Pillar root', self.config['pillar_root']
            )
            self.config['pillar_roots'] = {saltenv: self.config['pillar_root']}
        if self.config.get('file_root'):
            log.info(
                'Setting and using %s as the Salt file root', self.config['file_root']
            )
            self.config['file_root'] = {saltenv: self.config['file_root']}
        if self.config.get('installation_path'):
            salt.utils.stringutils.print_cli(curpath)
            return
        if self.config.get('display_file_roots'):
            salt.utils.stringutils.print_cli(
                'salt-sproxy is installed at: {}'.format(curpath)
            )
            salt.utils.stringutils.print_cli(
                '\nYou can configure the file_roots on the Master, e.g.,\n'
            )
            salt.utils.stringutils.print_cli(
                'file_roots:\n  {0}:\n    - {1}'.format(saltenv, curpath)
            )
            salt.utils.stringutils.print_cli('\n\nOr only for the Runners:\n')
            salt.utils.stringutils.print_cli(
                'runner_dirs:\n  - {}/_runners'.format(curpath)
            )
            return
        if self.config.get('save_file_roots'):
            updated = False
            with fopen(self.config['conf_file'], 'r+') as master_fp:
                master_cfg = safe_load(master_fp)
                if not master_cfg:
                    master_cfg = {}
                file_roots = master_cfg.get('file_roots', {saltenv: []}).get(
                    saltenv, []
                )
                runner_dirs = master_cfg.get('runner_dirs', [])
                sproxy_runners = os.path.join(curpath, '_runners')
                if curpath not in file_roots:
                    file_roots.append(curpath)
                    master_cfg['file_roots'] = {saltenv: file_roots}
                    updated = True
                    salt.utils.stringutils.print_cli(
                        '{} added to the file_roots:\n'.format(curpath)
                    )
                    salt.utils.stringutils.print_cli(
                        'file_roots:\n  {0}\n    - {1}\n'.format(
                            saltenv, '\n    -'.join(file_roots)
                        )
                    )
                if sproxy_runners not in runner_dirs:
                    runner_dirs.append(sproxy_runners)
                    master_cfg['runner_dirs'] = runner_dirs
                    updated = True
                    salt.utils.stringutils.print_cli(
                        '{} added to runner_dirs:\n'.format(sproxy_runners)
                    )
                    salt.utils.stringutils.print_cli(
                        'runner_dirs:\n  - {0}'.format('\n  - '.join(runner_dirs))
                    )
                if updated:
                    master_fp.seek(0)
                    safe_dump(master_cfg, master_fp, default_flow_style=False)
                    log.debug('Syncing Runners on the Master')
                    runner_client = salt.runner.RunnerClient(self.config)
                    sync_runners = runner_client.cmd(
                        'saltutil.sync_all',
                        kwarg={'saltenv': saltenv},
                        print_event=False,
                    )
                    log.debug('saltutil.sync_all output:')
                    log.debug(sync_runners)
                else:
                    salt.utils.stringutils.print_cli(
                        'The {} path is already included into the file_roots and runner_dirs'.format(
                            curpath
                        )
                    )
                salt.utils.stringutils.print_cli(
                    '\nNow you can start using salt-sproxy for '
                    'event-driven automation, and the Salt REST API.\n'
                    'See https://salt-sproxy.readthedocs.io/en/latest/salt_api.html'
                    '\nand https://salt-sproxy.readthedocs.io/en/latest/events.html '
                    'for more details.'
                )
            return
        # The code below executes the Runner sequence, but it swaps the function
        # to be invoked, and instead call ``napalm.execute``, passing the
        # function requested by the user from the CLI, as an argument.
        # The same goes with the CLI options that are sent as kwargs to the
        # proxy Runner.
        tgt = self.config['tgt']
        fun = self.config['fun']
        args = self.config['arg']
        kwargs = {}
        if 'output' not in self.config and fun in (
            'state.sls',
            'state.apply',
            'state.highstate',
        ):
            self.config['output'] = 'highstate'
        kwargs['progress'] = self.config.pop('progress', False)
        # To be able to reuse the proxy Runner (which is not yet available
        # natively in Salt), we can override the ``runner_dirs`` configuration
        # option to tell Salt to load that Runner too. This way, we can also
        # load other types of modules that may be required or we provide fixes
        # or backports - for example the Ansible Roster which doesn't work fine
        # pre Salt 2018.3 (in case anyone would like to use it).
        file_roots = self.config.get('file_roots', {saltenv: []})
        if saltenv not in file_roots:
            file_roots[saltenv] = []
        file_roots[saltenv].append(curpath)
        self.config['file_roots'] = file_roots
        runner_dirs = self.config.get('runner_dirs', [])
        runner_path = os.path.join(curpath, '_runners')
        runner_dirs.append(runner_path)
        self.config['runner_dirs'] = runner_dirs
        runner_client = None
        sync_all = self.config.get('sync_all', False)
        sync_grains = self.config.get('sync_grains', True)
        sync_modules = self.config.get('sync_modules', True)
        sync_roster = self.config.get('sync_roster', True)
        sync_proxy = self.config.get('sync_proxy', False)
        sync_executors = self.config.get('sync_executors', False)
        kwargs.update(
            {
                'sync_all': sync_all,
                'sync_roster': sync_roster,
                'sync_modules': sync_modules,
            }
        )
        if any(
            [
                sync_all,
                sync_grains,
                sync_modules,
                sync_roster,
                sync_proxy,
                sync_executors,
            ]
        ):
            runner_client = salt.runner.RunnerClient(self.config)
        if sync_all:
            log.debug('Sync all')
            sync_all_ret = runner_client.cmd(
                'saltutil.sync_all', kwarg={'saltenv': saltenv}, print_event=False
            )
            log.debug(sync_all_ret)
        if sync_grains and not sync_all:
            log.debug('Syncing grains')
            sync_grains_ret = runner_client.cmd(
                'saltutil.sync_grains',
                kwarg={
                    'saltenv': saltenv,
                    'extmod_whitelist': ','.join(
                        self.config.get('whitelist_grains', [])
                    ),
                    'extmod_blacklist': ','.join(self.config.get('disable_grains', [])),
                },
                print_event=False,
            )
            log.debug(sync_grains_ret)
        if self.config.get('module_dirs_cli'):
            log.debug(
                'Loading execution modules from the dirs provided via --module-dirs'
            )
            module_dirs = self.config.get('module_dirs', [])
            module_dirs.extend(self.config['module_dirs_cli'])
            self.config['module_dirs'] = module_dirs
        if sync_modules and not sync_all:
            # Don't sync modules by default
            log.debug('Syncing modules')
            module_dirs = self.config.get('module_dirs', [])
            module_path = os.path.join(curpath, '_modules')
            module_dirs.append(module_path)
            self.config['module_dirs'] = module_dirs
            # No need to explicitly load the modules here, as during runtime,
            # Salt is anyway going to load the modules on the fly.
            sync_modules_ret = runner_client.cmd(
                'saltutil.sync_modules',
                kwarg={
                    'saltenv': saltenv,
                    'extmod_whitelist': ','.join(
                        self.config.get('whitelist_modules', [])
                    ),
                    'extmod_blacklist': ','.join(
                        self.config.get('disable_modules', [])
                    ),
                },
                print_event=False,
            )
            log.debug(sync_modules_ret)
        # Resync Roster module to load the ones we have here in the library, and
        # potentially others provided by the user in their environment
        if sync_roster and not sync_all and self.config.get('roster'):
            # Sync Rosters by default
            log.debug('Syncing roster')
            roster_dirs = self.config.get('roster_dirs', [])
            roster_path = os.path.join(curpath, '_roster')
            roster_dirs.append(roster_path)
            self.config['roster_dirs'] = roster_dirs
            sync_roster_ret = runner_client.cmd(
                'saltutil.sync_roster',
                kwarg={'saltenv': saltenv, 'extmod_whitelist': self.config['roster']},
                print_event=False,
            )
            log.debug(sync_roster_ret)
        if sync_proxy and not sync_all:
            log.debug('Syncing Proxy modules')
            proxy_dirs = self.config.get('proxy_dirs', [])
            proxy_path = os.path.join(curpath, '_proxy')
            proxy_dirs.append(proxy_path)
            self.config['proxy_dirs'] = proxy_dirs
            sync_proxy_ret = runner_client.cmd(
                'saltutil.sync_proxymodules',
                kwarg={
                    'saltenv': saltenv,
                    'extmod_whitelist': ','.join(
                        self.config.get('whitelist_proxys', [])
                    ),
                    'extmod_blacklist': ','.join(self.config.get('disable_proxys', [])),
                },
                print_event=False,
            )
            log.debug(sync_proxy_ret)
        if sync_executors and not sync_all:
            log.debug('Syncing Executors modules')
            executor_dirs = self.config.get('executor_dirs', [])
            executor_path = os.path.join(curpath, '_executors')
            executor_dirs.append(executor_path)
            self.config['executor_dirs'] = executor_dirs
            sync_executors_ret = runner_client.cmd(
                'saltutil.sync_executors',
                kwarg={
                    'saltenv': saltenv,
                    'extmod_whitelist': ','.join(
                        self.config.get('whitelist_executors', [])
                    ),
                    'extmod_blacklist': ','.join(
                        self.config.get('disable_executors', [])
                    ),
                },
                print_event=False,
            )
            log.debug(sync_executors_ret)
        if self.config.get('states_dir'):
            states_dirs = self.config.get('states_dirs', [])
            states_dirs.append(self.config['states_dir'])
            self.config['states_dirs'] = states_dirs
        self.config['fun'] = 'proxy.execute'
        tmp_args = args[:]
        for index, arg in enumerate(tmp_args):
            if isinstance(arg, dict) and '__kwarg__' in arg:
                args.pop(index)
                kwargs = arg
        kwargs['__kwarg__'] = True
        tgt_types = (
            'compound',
            'list',
            'grain',
            'pcre',
            'grain_pcre',
            'pillar',
            'pillar_pcre',
            'pillar_target',
            'nodegroup',
        )
        kwargs['tgt_type'] = 'glob'
        for tgt_type in tgt_types:
            if hasattr(self.options, tgt_type) and getattr(self.options, tgt_type):
                kwargs['tgt_type'] = tgt_type
        kwargs_opts = (
            'preview_target',
            'batch_size',
            'batch_wait',
            'roster',
            'timeout',
            'static',
            'no_connect',
            'failhard',
            'summary',
            'verbose',
            'show_jid',
            'hide_timeout',
            'progress',
            'returner',
            'target_cache',
            'returner_config',
            'returner_kwargs',
        )
        for kwargs_opt in kwargs_opts:
            if getattr(self.options, kwargs_opt) is not None:
                kwargs[kwargs_opt] = getattr(self.options, kwargs_opt)
        reverse_opts = {
            # option_name: runner_kwarg
            'no_cached_grains': 'use_cached_grains',
            'no_cached_pillar': 'use_cached_pillar',
            'no_grains': 'with_grains',
            'no_pillar': 'with_pillar',
            'dont_cache_grains': 'cache_grains',
            'dont_cache_pillar': 'cache_pillar',
        }
        for opt, kwarg in six.iteritems(reverse_opts):
            if getattr(self.options, opt):
                kwargs[kwarg] = False
        kwargs['events'] = self.config.get('events', False)
        kwargs['use_existing_proxy'] = self.config.get('use_existing_proxy', False)
        kwargs['test_ping'] = self.config.get('test_ping', False)
        kwargs['target_cache_timeout'] = self.config.get(
            'target_cache_timeout', 60
        )  # seconds
        kwargs['args'] = args
        kwargs['default_grains'] = self.config.get(
            'sproxy_grains',
            self.config.get('default_grains', self.config.get('grains')),
        )
        kwargs['default_pillar'] = self.config.get(
            'sproxy_pillar',
            self.config.get('default_pillar', self.config.get('pillar')),
        )
        kwargs['preload_targeting'] = self.config.get('preload_targeting', False)
        kwargs['invasive_targeting'] = self.config.get('invasive_targeting', False)
        kwargs['failhard'] = self.config.get('failhard', False)
        self.config['arg'] = [tgt, fun, kwargs]
        runner = salt.runner.Runner(self.config)

        if self.config.get('doc', True):
            # late import as salt.loader adds up some execution time, and we
            # don't want that, but only when displaying docs.
            from salt.loader import utils, grains, minion_mods

            runner.opts['fun'] = fun
            runner.opts['grains'] = grains(runner.opts)
            runner._functions = minion_mods(runner.opts, utils=utils(runner.opts))

        # Run this here so SystemExit isn't raised anywhere else when
        # someone tries to use the runners via the python API
        try:
            if check_user(self.config['user']):
                pr = activate_profile(profiling_enabled)
                try:
                    ret = runner.run()
                    # In older versions ret['data']['retcode'] was used
                    # for signaling the return code. This has been
                    # changed for the orchestrate runner, but external
                    # runners might still use it. For this reason, we
                    # also check ret['data']['retcode'] if
                    # ret['retcode'] is not available.
                    if 'retcode' in runner.context:
                        self.exit(runner.context['retcode'])
                    if isinstance(ret, dict) and 'retcode' in ret:
                        self.exit(ret['retcode'])
                    elif isinstance(ret, dict) and 'retcode' in ret.get('data', {}):
                        self.exit(ret['data']['retcode'])
                finally:
                    output_profile(
                        pr, stats_path=self.options.profiling_path, stop=True
                    )

        except SaltClientError as exc:
            raise SystemExit from exc



# -*- coding: utf-8 -*-
'''
Various features for the Roster modules.
'''
from __future__ import absolute_import

import re
import fnmatch
import logging

import salt.cache
from salt.ext import six
import salt.utils.minions

import salt.utils.dictupdate
from salt.defaults import DEFAULT_TARGET_DELIM

try:
    from salt.utils.data import subdict_match
except ImportError:
    from salt.utils import subdict_match

log = logging.getLogger(__name__)


def load_cache(pool, __runner__, opts, tgt, tgt_type=None):
    '''
    Load the Pillar and Grain cache, as required, and merge the Roster Grains
    and Pillar into.
    '''
    if opts.get('grains'):
        for device, device_opts in six.iteritems(pool):
            if 'minion_opts' not in device_opts:
                device_opts['minion_opts'] = {}
            if 'grains' not in device_opts['minion_opts']:
                device_opts['minion_opts']['grains'] = {}
            device_opts['minion_opts']['grains'] = salt.utils.dictupdate.merge(
                opts['grains'], device_opts['minion_opts']['grains'], merge_lists=True,
            )
    if tgt_type in ('glob', 'pcre', 'list'):
        # When the target type is glob, pcre, or list, we don't require grains
        # or pillar loaded from the cache, because the targeting won't depend on
        # those.
        return pool
    if not opts.get('use_cached_grains', True) and not opts.get(
        'use_cached_pillar', True
    ):
        return pool
    # NOTE: It wouldn't be feasible to use the cache.grains or cache.pillar
    # Runners as they rely on fetching data from the Master, for Minions that
    # are accepted. What we're doing here is reading straight from the cache.
    log.debug('Loading cached and merging into the Roster data')
    cache = salt.cache.factory(opts)
    cache_pool = cache.list('minions')
    for device in cache_pool:
        if device not in pool:
            log.trace('%s has cache, but is not in the Roster pool', device)
            continue
        if 'minion_opts' not in pool[device]:
            pool[device]['minion_opts'] = {'grains': {}, 'pillar': {}}
        cache_key = 'minions/{}/data'.format(device)
        if opts.get('target_use_cached_grains', True) and tgt_type in (
            'compound',
            'grain',
            'grain_pcre',
            'nodegroup',
        ):
            log.debug('Fetching cached Grains for %s', device)
            cached_grains = cache.fetch(cache_key, 'grains')
            if cached_grains:
                pool[device]['minion_opts']['grains'] = salt.utils.dictupdate.merge(
                    cached_grains,
                    pool[device]['minion_opts'].get('grains', {}),
                    merge_lists=True,
                )
        if opts.get('target_use_cached_pillar', True) and tgt_type in (
            'compound',
            'pillar',
            'pillar_pcre',
            'pillar_target',
            'nodegroup',
        ):
            log.debug('Fetching cached Pillar for %s', device)
            cached_pillar = cache.fetch(cache_key, 'pillar')
            if cached_pillar:
                pool[device]['minion_opts']['pillar'] = salt.utils.dictupdate.merge(
                    cached_pillar,
                    pool[device]['minion_opts'].get('pillar', {}),
                    merge_lists=True,
                )
    log.debug('The device pool with the cached data')
    log.debug(pool)
    return pool


def glob(pool, tgt, opts=None):
    '''
    '''
    log.debug('Glob matching on %s ? %s', pool.items(), tgt)
    return {
        minion: pool[minion] for minion in pool.keys() if fnmatch.fnmatch(minion, tgt)
    }


def grain(pool, tgt, opts=None):
    '''
    '''
    delimiter = opts.get('delimiter', DEFAULT_TARGET_DELIM)
    log.debug('Grain matching on %s, over %s', tgt, pool)
    ret = {
        minion: pool[minion]
        for minion in pool.keys()
        if subdict_match(
            pool[minion].get('minion_opts', {}).get('grains', {}),
            tgt,
            delimiter=delimiter,
        )
    }
    log.debug('Grain match returned')
    log.debug(ret)
    return ret


def grain_pcre(pool, tgt, opts=None):
    '''
    '''
    delimiter = opts.get('delimiter', DEFAULT_TARGET_DELIM)
    log.debug('Grain PCRE matching on %s, over %s', tgt, pool)
    ret = {
        minion: pool[minion]
        for minion in pool.keys()
        if subdict_match(
            pool[minion].get('minion_opts', {}).get('grains', {}),
            tgt,
            delimiter=delimiter,
            regex_match=True,
        )
    }
    log.debug('Grain PCRE match returned')
    log.debug(ret)
    return ret


def pillar(pool, tgt, opts=None):
    '''
    '''
    delimiter = opts.get('delimiter', DEFAULT_TARGET_DELIM)
    log.debug('Pillar matching on %s, over %s', tgt, pool)
    ret = {
        minion: pool[minion]
        for minion in pool.keys()
        if subdict_match(
            pool[minion].get('minion_opts', {}).get('pillar', {}),
            tgt,
            delimiter=delimiter,
        )
    }
    log.debug('Pillar match returned')
    log.debug(ret)
    return ret


def pillar_pcre(pool, tgt, opts=None):
    '''
    '''
    delimiter = opts.get('delimiter', DEFAULT_TARGET_DELIM)
    log.debug('Pillar PCRE matching on %s, over %s', tgt, pool)
    ret = {
        minion: pool[minion]
        for minion in pool.keys()
        if subdict_match(
            pool[minion].get('minion_opts', {}).get('pillar', {}),
            tgt,
            delimiter=delimiter,
            regex_match=True,
        )
    }
    log.debug('Pillar PCRE match returned')
    log.debug(ret)
    return ret


def list_(pool, tgt, opts=None):
    '''
    '''
    log.debug('List matching on %s ? %s', pool.items(), tgt)
    return {minion: pool[minion] for minion in pool.keys() if minion in tgt}


def pcre(pool, tgt, opts=None):
    '''
    '''
    log.debug('PCRE matching on %s ? %s', pool.items(), tgt)
    rgx = re.compile(tgt)
    return {minion: pool[minion] for minion in pool.keys() if rgx.search(minion)}


def nodegroup(pool, tgt, opts=None):
    '''
    '''
    nodegroups = opts.get('nodegroups', {})
    # tgt is the name of the nodegroup
    if tgt not in nodegroups:
        return {}
    nodegroup_expr = nodegroups[tgt]
    return compound(pool, nodegroup_expr, opts=opts)


TGT_FUN = {
    'glob': glob,
    'G': grain,
    'grain': grain,
    'P': grain_pcre,
    'grain_pcre': grain_pcre,
    'I': pillar,
    'pillar': pillar,
    'pillar_target': pillar,
    'J': pillar_pcre,
    'pillar_pcre': pillar_pcre,
    'L': list_,
    'list': list_,
    'N': nodegroup,
    'nodegroup': nodegroup,
    'E': pcre,
    'pcre': pcre,
}


def compound(pool, tgt, opts=None):
    '''
    Execute a compound match on a pool of devices returned by the Roster. The
    Roster module must collect the entire list of devices managed by this Master
    / salt-sproxy instance, and this function helps filtering out the Minions
    that are outside of this target expression.
    This function returns the list of Minions matched by the target expression,
    together with their opts (i.e., extra Grains and Pillar).
    The first argument passed in is ``pool`` which is a dictionary containing
    the total group of devices that can possibly be managed, and their opts.
    '''
    minions = {}
    if not isinstance(tgt, six.string_types) and not isinstance(tgt, (list, tuple)):
        log.error('Compound target received that is neither string, list nor tuple')
        return minions

    results = []
    opers = ['and', 'or', 'not', '(', ')']

    if isinstance(tgt, six.string_types):
        words = tgt.split()
    else:
        words = tgt[:]

    while words:
        word = words.pop(0)
        target_info = salt.utils.minions.parse_target(word)

        if word in opers:
            if results:
                if results[-1] == '(' and word in ('and', 'or'):
                    log.error('Invalid beginning operator after "(": %s', word)
                    return {}
                if word == 'not':
                    if not results[-1] in ('and', 'or', '('):
                        results.append('and')
                results.append(word)
            else:
                # seq start with binary oper, fail
                if word not in ['(', 'not']:
                    log.error('Invalid beginning operator: %s', word)
                    return {}
                results.append(word)

        elif target_info and target_info['engine']:
            engine = TGT_FUN.get(target_info['engine'])
            if not engine:
                # If an unknown engine is called at any time, fail out
                log.error(
                    'Unrecognized target engine "%s" for target ' 'expression "%s"',
                    target_info['engine'],
                    word,
                )
                return {}
            res = engine(pool, target_info['pattern'], opts=opts)
            results.append(str(set(res.keys())))

        else:
            res = glob(pool, word, opts=opts)
            results.append(str(set(res.keys())))

    log.debug('Collected individual results')
    log.debug(results)

    expr_chunks = []
    parens_count = 0
    universe = set(pool.keys())
    # Building the target list, using set theory operations
    # `X and Y` becomes `X & Y`
    # `X or Y` becomes `X | Y`
    # `X and not Y` becomes `X - Y`
    # `X or not Y` becomes `X | ( U - Y )` where U is the universe, in our case
    #   being the total group of devices possibly being managed by this
    #   salt-sproxy instance.
    #
    # The following iteration goes through the list of results, and replaces the
    # CLI operational words, as per logic described above.
    # TODO: the `and not` operation is fine, but the `or not` might be weak as
    # it introduces another set of parens which may mess up the expression.
    # Below, when evaluating the expression, I've added a block to catch the
    # exception and ask for bug report
    for index, res in enumerate(results):
        if res == 'not':
            res = '{} -'.format(universe)
        if res == 'and':
            if results[index + 1] == 'not':
                res = '-'
                results[index + 1] = ''
            else:
                res = '&'
        elif res == 'or':
            if results[index + 1] == 'not':
                res = '| ( {} -'.format(universe)
                parens_count += 1
                results[index + 1] = ''
            else:
                res = '|'
        expr_chunks.append(res)
    expr_chunks += ')' * parens_count

    match_expr = ' '.join(expr_chunks)
    log.debug('Matching expression: %s', match_expr)
    try:
        matched_minions = eval(match_expr)  # pylint: disable=W0123
    except SyntaxError:
        log.error('Looks like this target expression is failing.')
        if 'or not' in tgt:
            log.error(
                'This may be a salt-sproxy bug, please report at: \n'
                'https://github.com/mirceaulinic/salt-sproxy/issues/new?'
                'labels=bug%2C+pending+triage&template=bug_report.md'
                '&title=Issue%20when%20using%20the%20compound%20target'
            )
        return {}
    log.debug('Matched Minions')
    log.debug(matched_minions)

    return {minion: pool[minion] for minion in matched_minions}


TGT_FUN['compound'] = compound



# -*- coding: utf-8 -*-
'''
NetBox
======

Module to query NetBox

:codeauthor: Zach Moody <zmoody@do.co>
:maturity:   new
:depends:    pynetbox

.. note::

    This code, distributed as part of ``salt-sproxy``, has been copied from the
    main Salt project, maintained by SaltStack, to provide various enhancements
    and fixes to the original module.

The following config should be in the minion config file. In order to
work with ``secrets`` you should provide a token and path to your
private key file:

.. code-block:: yaml

  netbox:
    url: <NETBOX_URL>
    token: <NETBOX_USERNAME_API_TOKEN (OPTIONAL)>
    keyfile: </PATH/TO/NETBOX/KEY (OPTIONAL)>

.. versionadded:: 2018.3.0

    This module has been introduced in Salt release 2018.3.0.

    In ``salt-sproxy``, this module has been included beginning with version
    2019.10.0.
'''
from __future__ import absolute_import, unicode_literals, print_function

import re
import logging

from salt.exceptions import CommandExecutionError

try:
    import pynetbox
    from pynetbox import RequestError

    HAS_PYNETBOX = True
except ImportError:
    HAS_PYNETBOX = False

log = logging.getLogger(__name__)

AUTH_ENDPOINTS = ('secrets',)

__func_alias__ = {'filter_': 'filter', 'get_': 'get'}


def __virtual__():
    '''
    pynetbox must be installed.
    '''
    if not HAS_PYNETBOX:
        return (
            False,
            'The netbox execution module cannot be loaded: '
            'pynetbox library is not installed.',
        )
    else:
        return True


def _config():
    config = __salt__['config.get']('netbox')
    if not config:
        raise CommandExecutionError(
            'NetBox execution module configuration could not be found'
        )
    return config


def _nb_obj(auth_required=False):
    pynb_kwargs = {}
    pynb_kwargs['token'] = _config().get('token')
    if auth_required:
        pynb_kwargs['private_key_file'] = _config().get('keyfile')
    return pynetbox.api(_config().get('url'), **pynb_kwargs)


def _strip_url_field(input_dict):
    if 'url' in input_dict.keys():
        del input_dict['url']
    for k, v in input_dict.items():
        if isinstance(v, dict):
            _strip_url_field(v)
    return input_dict


def _dict(iterable):
    if iterable:
        return dict(iterable)
    else:
        return {}


def _add(app, endpoint, payload):
    '''
    POST a payload
    '''
    nb = _nb_obj(auth_required=True)
    try:
        return getattr(getattr(nb, app), endpoint).create(**payload)
    except RequestError as e:  # pylint: disable=undefined-variable
        log.error('%s, %s, %s', e.req.request.headers, e.request_body, e.error)
        return False


def slugify(value):
    ''''
    Slugify given value.
    Credit to Djangoproject https://docs.djangoproject.com/en/2.0/_modules/django/utils/text/#slugify
    '''
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[-\s]+', '-', value)


def _get(app, endpoint, id=None, auth_required=False, **kwargs):
    '''
    Helper function to do a GET request to Netbox.
    Returns the actual pynetbox object, which allows manipulation from other functions.
    '''
    nb = _nb_obj(auth_required=auth_required)
    if id:
        item = getattr(getattr(nb, app), endpoint).get(id)
    else:
        kwargs = __utils__['args.clean_kwargs'](**kwargs)
        item = getattr(getattr(nb, app), endpoint).get(**kwargs)
    return item


def _if_name_unit(if_name):
    if_name_split = if_name.split('.')
    if len(if_name_split) == 2:
        return if_name_split
    return if_name, '0'


def filter_(app, endpoint, **kwargs):
    '''
    Get a list of items from NetBox.

    app
        String of netbox app, e.g., ``dcim``, ``circuits``, ``ipam``
    endpoint
        String of app endpoint, e.g., ``sites``, ``regions``, ``devices``
    kwargs
        Optional arguments that can be used to filter.
        All filter keywords are available in Netbox,
        which can be found by surfing to the corresponding API endpoint,
        and clicking Filters. e.g., ``role=router``

    Returns a list of dictionaries

    .. code-block:: bash

        salt myminion netbox.filter dcim devices status=1 role=router
    '''
    ret = []
    nb = _nb_obj(auth_required=True if app in AUTH_ENDPOINTS else False)
    clean_kwargs = __utils__['args.clean_kwargs'](**kwargs)
    if not clean_kwargs:
        nb_query = getattr(getattr(nb, app), endpoint).all()
    else:
        nb_query = getattr(getattr(nb, app), endpoint).filter(**clean_kwargs)
    if nb_query:
        ret = [_strip_url_field(dict(i)) for i in nb_query]
    return ret


def get_(app, endpoint, id=None, **kwargs):
    '''
    Get a single item from NetBox.

    app
        String of netbox app, e.g., ``dcim``, ``circuits``, ``ipam``
    endpoint
        String of app endpoint, e.g., ``sites``, ``regions``, ``devices``

    Returns a single dictionary

    To get an item based on ID.

    .. code-block:: bash

        salt myminion netbox.get dcim devices id=123

    Or using named arguments that correspond with accepted filters on
    the NetBox endpoint.

    .. code-block:: bash

        salt myminion netbox.get dcim devices name=my-router
    '''
    return _dict(
        _get(
            app,
            endpoint,
            id=id,
            auth_required=True if app in AUTH_ENDPOINTS else False,
            **kwargs
        )
    )


def create_manufacturer(name):
    '''
    .. versionadded:: 2019.2.0

    Create a device manufacturer.

    name
        The name of the manufacturer, e.g., ``Juniper``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_manufacturer Juniper
    '''
    nb_man = get_('dcim', 'manufacturers', name=name)
    if nb_man:
        return False
    else:
        payload = {'name': name, 'slug': slugify(name)}
        man = _add('dcim', 'manufacturers', payload)
        if man:
            return {'dcim': {'manufacturers': payload}}
        else:
            return False


def create_device_type(model, manufacturer):
    '''
    .. versionadded:: 2019.2.0

    Create a device type. If the manufacturer doesn't exist, create a new manufacturer.

    model
        String of device model, e.g., ``MX480``
    manufacturer
        String of device manufacturer, e.g., ``Juniper``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_device_type MX480 Juniper
    '''
    nb_type = get_('dcim', 'device-types', model=model)
    if nb_type:
        return False
    nb_man = get_('dcim', 'manufacturers', name=manufacturer)
    new_man = None
    if not nb_man:
        new_man = create_manufacturer(manufacturer)
    payload = {'model': model, 'manufacturer': nb_man['id'], 'slug': slugify(model)}
    typ = _add('dcim', 'device-types', payload)
    ret_dict = {'dcim': {'device-types': payload}}
    if new_man:
        ret_dict['dcim'].update(new_man['dcim'])
    if typ:
        return ret_dict
    else:
        return False


def create_device_role(role, color):
    '''
    .. versionadded:: 2019.2.0

    Create a device role

    role
        String of device role, e.g., ``router``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_device_role router
    '''
    nb_role = get_('dcim', 'device-roles', name=role)
    if nb_role:
        return False
    else:
        payload = {'name': role, 'slug': slugify(role), 'color': color}
        role = _add('dcim', 'device-roles', payload)
        if role:
            return {'dcim': {'device-roles': payload}}
        else:
            return False


def create_platform(platform):
    '''
    .. versionadded:: 2019.2.0

    Create a new device platform

    platform
        String of device platform, e.g., ``junos``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_platform junos
    '''
    nb_platform = get_('dcim', 'platforms', slug=slugify(platform))
    if nb_platform:
        return False
    else:
        payload = {'name': platform, 'slug': slugify(platform)}
        plat = _add('dcim', 'platforms', payload)
        if plat:
            return {'dcim': {'platforms': payload}}
        else:
            return False


def create_site(site):
    '''
    .. versionadded:: 2019.2.0

    Create a new device site

    site
        String of device site, e.g., ``BRU``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_site BRU
    '''
    nb_site = get_('dcim', 'sites', name=site)
    if nb_site:
        return False
    else:
        payload = {'name': site, 'slug': slugify(site)}
        site = _add('dcim', 'sites', payload)
        if site:
            return {'dcim': {'sites': payload}}
        else:
            return False


def create_device(name, role, model, manufacturer, site):
    '''
    .. versionadded:: 2019.2.0

    Create a new device with a name, role, model, manufacturer and site.
    All these components need to be already in Netbox.

    name
        The name of the device, e.g., ``edge_router``
    role
        String of device role, e.g., ``router``
    model
        String of device model, e.g., ``MX480``
    manufacturer
        String of device manufacturer, e.g., ``Juniper``
    site
        String of device site, e.g., ``BRU``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_device edge_router router MX480 Juniper BRU
    '''
    try:
        nb_role = get_('dcim', 'device-roles', name=role)
        if not nb_role:
            return False

        nb_type = get_('dcim', 'device-types', model=model)
        if not nb_type:
            return False
        nb_site = get_('dcim', 'sites', name=site)
        if not nb_site:
            return False

        status = {'label': "Active", 'value': 1}
    except RequestError as e:  # pylint: disable=undefined-variable
        log.error('%s, %s, %s', e.req.request.headers, e.request_body, e.error)
        return False

    payload = {
        'name': name,
        'display_name': name,
        'slug': slugify(name),
        'device_type': nb_type['id'],
        'device_role': nb_role['id'],
        'site': nb_site['id'],
    }
    new_dev = _add('dcim', 'devices', payload)
    if new_dev:
        return {'dcim': {'devices': payload}}
    else:
        return False


def update_device(name, **kwargs):
    '''
    .. versionadded:: 2019.2.0

    Add attributes to an existing device, identified by name.

    name
        The name of the device, e.g., ``edge_router``
    kwargs
       Arguments to change in device, e.g., ``serial=JN2932930``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.update_device edge_router serial=JN2932920
    '''
    kwargs = __utils__['args.clean_kwargs'](**kwargs)
    nb_device = _get('dcim', 'devices', auth_required=True, name=name)
    for k, v in kwargs.items():
        setattr(nb_device, k, v)
    try:
        nb_device.save()
        return {'dcim': {'devices': kwargs}}
    except RequestError as e:  # pylint: disable=undefined-variable
        log.error('%s, %s, %s', e.req.request.headers, e.request_body, e.error)
        return False


def create_inventory_item(
    device_name,
    item_name,
    manufacturer_name=None,
    serial='',
    part_id='',
    description='',
):
    '''
    .. versionadded:: 2019.2.0

    Add an inventory item to an existing device.

    device_name
        The name of the device, e.g., ``edge_router``.
    item_name
        String of inventory item name, e.g., ``Transceiver``.

    manufacturer_name
        String of inventory item manufacturer, e.g., ``Fiberstore``.

    serial
        String of inventory item serial, e.g., ``FS1238931``.

    part_id
        String of inventory item part id, e.g., ``740-01234``.

    description
        String of inventory item description, e.g., ``SFP+-10G-LR``.

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_inventory_item edge_router Transceiver part_id=740-01234
    '''
    nb_device = get_('dcim', 'devices', name=device_name)
    if not nb_device:
        return False
    if manufacturer_name:
        nb_man = get_('dcim', 'manufacturers', name=manufacturer_name)
        if not nb_man:
            create_manufacturer(manufacturer_name)
            nb_man = get_('dcim', 'manufacturers', name=manufacturer_name)
    payload = {
        'device': nb_device['id'],
        'name': item_name,
        'description': description,
        'serial': serial,
        'part_id': part_id,
        'parent': None,
    }
    if manufacturer_name:
        payload['manufacturer'] = nb_man['id']
    done = _add('dcim', 'inventory-items', payload)
    if done:
        return {'dcim': {'inventory-items': payload}}
    else:
        return done


def delete_inventory_item(item_id):
    '''
    .. versionadded:: 2019.2.0

    Remove an item from a devices inventory. Identified by the netbox id

    item_id
        Integer of item to be deleted

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.delete_inventory_item 1354
    '''
    nb_inventory_item = _get('dcim', 'inventory-items', auth_required=True, id=item_id)
    nb_inventory_item.delete()
    return {'DELETE': {'dcim': {'inventory-items': item_id}}}


def create_interface_connection(interface_a, interface_b):
    '''
    .. versionadded:: 2019.2.0

    Create an interface connection between 2 interfaces

    interface_a
        Interface id for Side A
    interface_b
        Interface id for Side B

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_interface_connection 123 456
    '''
    payload = {'interface_a': interface_a, 'interface_b': interface_b}
    ret = _add('dcim', 'interface-connections', payload)
    if ret:
        return {'dcim': {'interface-connections': {ret['id']: payload}}}
    else:
        return ret


def get_interfaces(device_name=None, **kwargs):
    '''
    .. versionadded:: 2019.2.0

    Returns interfaces for a specific device using arbitrary netbox filters

    device_name
        The name of the device, e.g., ``edge_router``
    kwargs
        Optional arguments to be used for filtering

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.get_interfaces edge_router name="et-0/0/5"

    '''
    if not device_name:
        device_name = __opts__['id']
    netbox_device = get_('dcim', 'devices', name=device_name)
    return filter_('dcim', 'interfaces', device_id=netbox_device['id'], **kwargs)


def openconfig_interfaces(device_name=None):
    '''
    .. versionadded:: 2019.2.0

    Return a dictionary structured as standardised in the
    `openconfig-interfaces <http://ops.openconfig.net/branches/master/openconfig-interfaces.html>`_
    YANG model, containing physical and configuration data available in Netbox,
    e.g., IP addresses, MTU, enabled / disabled, etc.

    device_name: ``None``
        The name of the device to query the interface data for. If not provided,
        will use the Minion ID.

    CLI Example:

    .. code-block:: bash

        salt '*' netbox.openconfig_interfaces
        salt '*' netbox.openconfig_interfaces device_name=cr1.thn.lon
    '''
    oc_if = {}
    interfaces = get_interfaces(device_name=device_name)
    ipaddresses = get_ipaddresses(device_name=device_name)
    for interface in interfaces:
        if_name, if_unit = _if_name_unit(interface['name'])
        if if_name not in oc_if:
            oc_if[if_name] = {
                'config': {'name': if_name},
                'subinterfaces': {'subinterface': {}},
            }
        if if_unit == '0':
            oc_if[if_name]['config']['enabled'] = interface['enabled']
            if interface['description']:
                if if_name == interface['name']:
                    # When that's a real unit 0 interface
                    # Otherwise it will inherit the description from the subif
                    oc_if[if_name]['config']['description'] = str(
                        interface['description']
                    )
                else:
                    subif_descr = {
                        'subinterfaces': {
                            'subinterface': {
                                if_unit: {
                                    'config': {
                                        'description': str(interface['description'])
                                    }
                                }
                            }
                        }
                    }
                    oc_if[if_name] = __utils__['dictupdate.update'](
                        oc_if[if_name], subif_descr
                    )
            if interface['mtu']:
                oc_if[if_name]['config']['mtu'] = int(interface['mtu'])
        else:
            oc_if[if_name]['subinterfaces']['subinterface'][if_unit] = {
                'config': {'index': int(if_unit), 'enabled': interface['enabled']}
            }
            if interface['description']:
                oc_if[if_name]['subinterfaces']['subinterface'][if_unit]['config'][
                    'description'
                ] = str(interface['description'])
    for ipaddress in ipaddresses:
        ip, prefix_length = ipaddress['address'].split('/')
        if_name = ipaddress['interface']['name']
        if_name, if_unit = _if_name_unit(if_name)
        ipvkey = 'ipv{}'.format(ipaddress['family'])
        if if_unit not in oc_if[if_name]['subinterfaces']['subinterface']:
            oc_if[if_name]['subinterfaces']['subinterface'][if_unit] = {
                'config': {'index': int(if_unit), 'enabled': True}
            }
        if ipvkey not in oc_if[if_name]['subinterfaces']['subinterface'][if_unit]:
            oc_if[if_name]['subinterfaces']['subinterface'][if_unit][ipvkey] = {
                'addresses': {'address': {}}
            }
        oc_if[if_name]['subinterfaces']['subinterface'][if_unit][ipvkey]['addresses'][
            'address'
        ][ip] = {'config': {'ip': ip, 'prefix_length': int(prefix_length)}}
    return {'interfaces': {'interface': oc_if}}


def openconfig_lacp(device_name=None):
    '''
    .. versionadded:: 2019.2.0

    Return a dictionary structured as standardised in the
    `openconfig-lacp <http://ops.openconfig.net/branches/master/openconfig-lacp.html>`_
    YANG model, with configuration data for Link Aggregation Control Protocol
    (LACP) for aggregate interfaces.

    .. note::
        The ``interval`` and ``lacp_mode`` keys have the values set as ``SLOW``
        and ``ACTIVE`` respectively, as this data is not currently available
        in Netbox, therefore defaulting to the values defined in the standard.
        See `interval <http://ops.openconfig.net/branches/master/docs/openconfig-lacp.html#lacp-interfaces-interface-config-interval>`_
        and `lacp-mode <http://ops.openconfig.net/branches/master/docs/openconfig-lacp.html#lacp-interfaces-interface-config-lacp-mode>`_
        for further details.

    device_name: ``None``
        The name of the device to query the LACP information for. If not provided,
        will use the Minion ID.

    CLI Example:

    .. code-block:: bash

        salt '*' netbox.openconfig_lacp
        salt '*' netbox.openconfig_lacp device_name=cr1.thn.lon
    '''
    oc_lacp = {}
    interfaces = get_interfaces(device_name=device_name)
    for interface in interfaces:
        if not interface['lag']:
            continue
        if_name, if_unit = _if_name_unit(interface['name'])
        parent_if = interface['lag']['name']
        if parent_if not in oc_lacp:
            oc_lacp[parent_if] = {
                'config': {
                    'name': parent_if,
                    'interval': 'SLOW',
                    'lacp_mode': 'ACTIVE',
                },
                'members': {'member': {}},
            }
        oc_lacp[parent_if]['members']['member'][if_name] = {}
    return {'lacp': {'interfaces': {'interface': oc_lacp}}}


def create_interface(
    device_name,
    interface_name,
    mac_address=None,
    description=None,
    enabled=None,
    lag=None,
    lag_parent=None,
    form_factor=None,
):
    '''
    .. versionadded:: 2019.2.0

    Attach an interface to a device. If not all arguments are provided,
    they will default to Netbox defaults.

    device_name
        The name of the device, e.g., ``edge_router``
    interface_name
        The name of the interface, e.g., ``TenGigE0/0/0/0``
    mac_address
        String of mac address, e.g., ``50:87:89:73:92:C8``
    description
        String of interface description, e.g., ``NTT``
    enabled
        String of boolean interface status, e.g., ``True``
    lag:
        Boolean of interface lag status, e.g., ``True``
    lag_parent
        String of interface lag parent name, e.g., ``ae13``
    form_factor
        Integer of form factor id, obtained through _choices API endpoint, e.g., ``200``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_interface edge_router ae13 description="Core uplink"
    '''
    nb_device = get_('dcim', 'devices', name=device_name)
    if not nb_device:
        return False
    if lag_parent:
        lag_interface = get_(
            'dcim', 'interfaces', device_id=nb_device['id'], name=lag_parent
        )
        if not lag_interface:
            return False
    if not description:
        description = ''
    if not enabled:
        enabled = 'false'
    # Set default form factor to 1200. This maps to SFP+ (10GE). This should be addressed by
    # the _choices endpoint.
    payload = {
        'device': nb_device['id'],
        'name': interface_name,
        'description': description,
        'enabled': enabled,
        'form_factor': 1200,
    }
    if form_factor is not None:
        payload['form_factor'] = form_factor
    if lag:
        payload['form_factor'] = 200
    if lag_parent:
        payload['lag'] = lag_interface['id']
    if mac_address:
        payload['mac_address'] = mac_address
    nb_interface = get_(
        'dcim', 'interfaces', device_id=nb_device['id'], name=interface_name
    )
    if not nb_interface:
        nb_interface = _add('dcim', 'interfaces', payload)
    if nb_interface:
        return {'dcim': {'interfaces': {nb_interface['id']: payload}}}
    else:
        return nb_interface


def update_interface(device_name, interface_name, **kwargs):
    '''
    .. versionadded:: 2019.2.0

    Update an existing interface with new attributes.

    device_name
        The name of the device, e.g., ``edge_router``
    interface_name
        The name of the interface, e.g., ``ae13``
    kwargs
        Arguments to change in interface, e.g., ``mac_address=50:87:69:53:32:D0``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.update_interface edge_router ae13 mac_address=50:87:69:53:32:D0
    '''
    nb_device = get_('dcim', 'devices', name=device_name)
    nb_interface = _get(
        'dcim',
        'interfaces',
        auth_required=True,
        device_id=nb_device['id'],
        name=interface_name,
    )
    if not nb_device:
        return False
    if not nb_interface:
        return False
    else:
        for k, v in __utils__['args.clean_kwargs'](**kwargs).items():
            setattr(nb_interface, k, v)
        try:
            nb_interface.save()
            return {'dcim': {'interfaces': {nb_interface.id: dict(nb_interface)}}}
        except RequestError as e:  # pylint: disable=undefined-variable
            log.error('%s, %s, %s', e.req.request.headers, e.request_body, e.error)
            return False


def delete_interface(device_name, interface_name):
    '''
    .. versionadded:: 2019.2.0

    Delete an interface from a device.

    device_name
        The name of the device, e.g., ``edge_router``.

    interface_name
        The name of the interface, e.g., ``ae13``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.delete_interface edge_router ae13
    '''
    nb_device = get_('dcim', 'devices', name=device_name)
    nb_interface = _get(
        'dcim',
        'interfaces',
        auth_required=True,
        device_id=nb_device['id'],
        name=interface_name,
    )
    if nb_interface:
        nb_interface.delete()
        return {
            'DELETE': {'dcim': {'interfaces': {nb_interface.id: nb_interface.name}}}
        }
    return False


def make_interface_lag(device_name, interface_name):
    '''
    .. versionadded:: 2019.2.0

    Update an interface to be a LAG.

    device_name
        The name of the device, e.g., ``edge_router``.

    interface_name
        The name of the interface, e.g., ``ae13``.

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.make_interface_lag edge_router ae13
    '''
    return update_interface(device_name, interface_name, form_factor=200)


def make_interface_child(device_name, interface_name, parent_name):
    '''
    .. versionadded:: 2019.2.0

    Set an interface as part of a LAG.

    device_name
        The name of the device, e.g., ``edge_router``.

    interface_name
        The name of the interface to be attached to LAG, e.g., ``xe-1/0/2``.

    parent_name
        The name of the LAG interface, e.g., ``ae13``.

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.make_interface_child xe-1/0/2 ae13
    '''
    nb_device = get_('dcim', 'devices', name=device_name)
    nb_parent = get_('dcim', 'interfaces', device_id=nb_device['id'], name=parent_name)
    if nb_device and nb_parent:
        return update_interface(device_name, interface_name, lag=nb_parent['id'])
    else:
        return False


def get_ipaddresses(device_name=None, **kwargs):
    '''
    .. versionadded:: 2019.2.0

    Filters for an IP address using specified filters

    device_name
        The name of the device to check for the IP address
    kwargs
        Optional arguments that can be used to filter, e.g., ``family=4``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.get_ipaddresses device_name family=4
    '''
    if not device_name:
        device_name = __opts__['id']
    netbox_device = get_('dcim', 'devices', name=device_name)
    return filter_('ipam', 'ip-addresses', device_id=netbox_device['id'], **kwargs)


def create_ipaddress(ip_address, family, device=None, interface=None):
    '''
    .. versionadded:: 2019.2.0

    Add an IP address, and optionally attach it to an interface.

    ip_address
        The IP address and CIDR, e.g., ``192.168.1.1/24``
    family
        Integer of IP family, e.g., ``4``
    device
        The name of the device to attach IP to, e.g., ``edge_router``
    interface
        The name of the interface to attach IP to, e.g., ``ae13``

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_ipaddress 192.168.1.1/24 4 device=edge_router interface=ae13
    '''
    nb_addr = None
    payload = {'family': family, 'address': ip_address}
    if interface and device:
        nb_device = get_('dcim', 'devices', name=device)
        if not nb_device:
            return False
        nb_interface = get_(
            'dcim', 'interfaces', device_id=nb_device['id'], name=interface
        )
        if not nb_interface:
            return False
        nb_addr = get_(
            'ipam',
            'ip-addresses',
            q=ip_address,
            interface_id=nb_interface['id'],
            family=family,
        )
        if nb_addr:
            log.error(nb_addr)
            return False
        else:
            payload['interface'] = nb_interface['id']
    ipaddr = _add('ipam', 'ip-addresses', payload)
    if ipaddr:
        return {'ipam': {'ip-addresses': payload}}
    else:
        return ipaddr


def delete_ipaddress(ipaddr_id):
    '''
    .. versionadded:: 2019.2.0

    Delete an IP address. IP addresses in Netbox are a combination of address
    and the interface it is assigned to.

    id
        The Netbox id for the IP address.

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.delete_ipaddress 9002
    '''

    nb_ipaddr = _get('ipam', 'ip-addresses', auth_required=True, id=ipaddr_id)
    if nb_ipaddr:
        nb_ipaddr.delete()
        return {'DELETE': {'ipam': {'ip-address': ipaddr_id}}}
    return False


def create_circuit_provider(name, asn=None):
    '''
    .. versionadded:: 2019.2.0

    Create a new Netbox circuit provider

    name
        The name of the circuit provider
    asn
        The ASN of the circuit provider

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_circuit_provider Telia 1299
    '''

    nb_circuit_provider = get_('circuits', 'providers', name=name)
    payload = {}
    if nb_circuit_provider:
        if nb_circuit_provider['asn'] == asn:
            return False
        else:
            log.error('Duplicate provider with different ASN: %s: %s', name, asn)
            raise CommandExecutionError(
                'Duplicate provider with different ASN: {}: {}'.format(name, asn)
            )
    else:
        payload = {'name': name, 'slug': slugify(name)}
        if asn:
            payload['asn'] = asn
        circuit_provider = _add('circuits', 'providers', payload)
        if circuit_provider:
            return {'circuits': {'providers': {circuit_provider['id']: payload}}}
        else:
            return circuit_provider


def get_circuit_provider(name, asn=None):
    '''
    .. versionadded:: 2019.2.0

    Get a circuit provider with a given name and optional ASN.

    name
        The name of the circuit provider
    asn
        The ASN of the circuit provider

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.get_circuit_provider Telia 1299
    '''
    if asn:
        nb_circuit_provider = get_('circuits', 'providers', asn=asn)
    else:
        nb_circuit_provider = get_('circuits', 'providers', name=name)
    return nb_circuit_provider


def create_circuit_type(name):
    '''
    .. versionadded:: 2019.2.0

    Create a new Netbox circuit type.

    name
        The name of the circuit type

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_circuit_type Transit
    '''
    nb_circuit_type = get_('circuits', 'circuit-types', slug=slugify(name))
    if nb_circuit_type:
        return False
    else:
        payload = {'name': name, 'slug': slugify(name)}
        circuit_type = _add('circuits', 'circuit-types', payload)
        if circuit_type:
            return {'circuits': {'circuit-types': {circuit_type['id']: payload}}}
        else:
            return circuit_type


def create_circuit(name, provider_id, circuit_type, description=None):
    '''
    .. versionadded:: 2019.2.0

    Create a new Netbox circuit

    name
        Name of the circuit
    provider_id
        The netbox id of the circuit provider
    circuit_type
        The name of the circuit type
    asn
        The ASN of the circuit provider
    description
        The description of the circuit

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_circuit NEW_CIRCUIT_01 Telia Transit 1299 "New Telia circuit"
    '''

    nb_circuit_provider = get_('circuits', 'providers', provider_id)
    nb_circuit_type = get_('circuits', 'circuit-types', slug=slugify(circuit_type))
    if nb_circuit_provider and nb_circuit_type:
        payload = {
            'cid': name,
            'provider': nb_circuit_provider['id'],
            'type': nb_circuit_type['id'],
        }
        if description:
            payload['description'] = description
        nb_circuit = get_('circuits', 'circuits', cid=name)
        if nb_circuit:
            return False
        circuit = _add('circuits', 'circuits', payload)
        if circuit:
            return {'circuits': {'circuits': {circuit['id']: payload}}}
        else:
            return circuit
    else:
        return False


def create_circuit_termination(
    circuit, interface, device, speed, xconnect_id=None, term_side='A'
):
    '''
    .. versionadded:: 2019.2.0

    Terminate a circuit on an interface

    circuit
        The name of the circuit
    interface
        The name of the interface to terminate on
    device
        The name of the device the interface belongs to
    speed
        The speed of the circuit, in Kbps
    xconnect_id
        The cross-connect identifier
    term_side
        The side of the circuit termination

    CLI Example:

    .. code-block:: bash

        salt myminion netbox.create_circuit_termination NEW_CIRCUIT_01 xe-0/0/1 myminion 10000 xconnect_id=XCON01
    '''

    nb_device = get_('dcim', 'devices', name=device)
    nb_interface = get_('dcim', 'interfaces', device_id=nb_device['id'], name=interface)
    nb_circuit = get_('circuits', 'circuits', cid=circuit)
    if nb_circuit and nb_device:
        nb_termination = get_('circuits', 'circuit-terminations', q=nb_circuit['cid'])
        if nb_termination:
            return False
        payload = {
            'circuit': nb_circuit['id'],
            'interface': nb_interface['id'],
            'site': nb_device['site']['id'],
            'port_speed': speed,
            'term_side': term_side,
        }
        if xconnect_id:
            payload['xconnect_id'] = xconnect_id
        circuit_termination = _add('circuits', 'circuit-terminations', payload)
        if circuit_termination:
            return {
                'circuits': {
                    'circuit-terminations': {circuit_termination['id']: payload}
                }
            }
        else:
            return circuit_termination



# -*- coding: utf-8 -*-
'''
SSH Executor module
===================

Used in conjunction with the SSH Proxy, to invoke Salt functions through Salt
thin, on a remove mchine accessed via SSH.
'''
from __future__ import absolute_import, unicode_literals

import logging

__virtualname__ = 'ssh'
__proxyenabled__ = ['ssh']


log = logging.getLogger(__name__)


def __virtual__():
    if 'proxy' not in __opts__:
        return False, 'SSH Executor is only meant to be used with SSH Proxy Minions'
    if __opts__.get('proxy', {}).get('proxytype') != __virtualname__:
        return False, 'Proxytype does not match: {0}'.format(__virtualname__)
    return True


def execute(opts, data, func, args, kwargs):
    '''
    Directly calls the given function with arguments
    '''
    if data['fun'] == 'saltutil.find_job':
        return __executors__['direct_call.execute'](opts, data, func, args, kwargs)
    return __proxy__['ssh.call'](data['fun'], *args, **kwargs)


def allow_missing_func(function):  # pylint: disable=unused-argument
    '''
    Allow all calls to be passed through to docker container.

    The ssh call will use direct_call, which will return back if the module
    was unable to be run.
    '''
    return True
